"""Context managers for common SCPI tasks."""

from .Instrument import Instrument
from .InstrumentErrors import assert_no_instrument_status_errors, TimeoutException, StatusException
from typing import List, Tuple
from pyvisa import VisaIOError


class InstrErrorSuppressor:
	"""Context-manager class to suppress exception coming from the instrument's status system.
	Other exceptions types are still raised.
	On entering the context, this class clears all the instrument status errors.
	:param io: RsInstrument session instance.
	:param visa_tout_ms: VISA Timeout in milliseconds, that is set for this context.
		Afterward, it is changed back. Default value: do-not-change.
	:param suppress_only_codes: You can enter a code or list of codes for errors to be suppressed.
		Other errors will be reported. Example: If you enter -113 here, only the 'Undefined Header' error will be suppressed.
		Default value: suppress-all-errors."""

	def __init__(self, io: Instrument, visa_tout_ms: int = 0, suppress_only_codes: int | List[int] = None):
		self._io: Instrument = io
		self._old_query_instr_status: bool = False
		self._old_visa_tout_ms: int = 0
		self._errors: List = []
		self._suppress_only_codes = suppress_only_codes
		self._exited: bool = False

		if visa_tout_ms > 0:
			self._old_visa_tout_ms = io.visa_timeout
			io.visa_timeout = visa_tout_ms

	def __enter__(self) -> 'InstrErrorSuppressor':
		"""Stuff to do when entering the context.
		Remember the old instrument status checking setting."""
		self._old_query_instr_status = self._io.query_instr_status
		self._io.query_instr_status = False
		self._io.clear_status()
		return self

	def __exit__(self, exc_type, value, traceback):
		"""Stuff to do when leaving the context.
				- Reads all the errors from the error queue.
				- Sets the Instrument Status Checking flag to the original value.
				- Sets the VISA Timeout to the original value.
			:returns: True, if the exception in the block should be suppressed."""
		self._io.query_instr_status = self._old_query_instr_status

		if self._old_visa_tout_ms > 0:
			self._io.visa_timeout = self._old_visa_tout_ms

		self.get_errors_occurred()
		self._exited = True

		# Even if the instrument status was switched OFF,
		#  the StatusException can happen when you query an invalid command.
		if isinstance(value, StatusException):
			suppressed = self._add_or_suppress_new_errors(value.errors_list)
			if suppressed:
				self._errors.extend(suppressed)
				return True  # Suppress this exception

		return False

	def get_errors_occurred(self) -> bool:
		"""Inside the context: Returns True, if any new errors occurred since the last call of this method.
		Outside the context: Returns True, if the accumulated error list contains errors."""
		if self._exited:
			return len(self._errors) > 0

		new_errors = self._io.query_all_syst_errors()
		suppressed = self._add_or_suppress_new_errors(new_errors)

		if suppressed:
			self._errors.extend(suppressed)
			return True

		return False

	def get_all_errors(self, incl_codes: bool = False) -> List[str] | List[Tuple[int, str]] | None:
		"""Returns all the errors accumulated during the call of this context.
		If no errors occurred, the method returns 'None'."""
		self.get_errors_occurred()
		if len(self._errors) == 0:
			return None
		if incl_codes:
			return self._errors
		return [x[1] for x in self._errors]

	def _add_or_suppress_new_errors(self, errors: List[Tuple[int, str]]) -> List[Tuple[int, str]]:
		"""Internal method, that takes the entered list of errors and checks whether they should be suppressed.
		Suppressed errors are added to the self._errors, and not suppresses are raised as StatusException."""
		if not errors:
			return errors

		if not self._suppress_only_codes:
			return errors

		# For cases where the _suppress_only_codes contain single value or list of values,
		# we have to suppress selectively.
		report: List[Tuple[int, str]] = []
		suppressed: List[Tuple[int, str]] = []
		for err in errors:
			suppress = False
			if isinstance(self._suppress_only_codes, int):
				suppress = err[0] == self._suppress_only_codes
			elif isinstance(self._suppress_only_codes, list):
				suppress = err[0] in self._suppress_only_codes
			if suppress:
				suppressed.append(err)
			else:
				# The error code should not be suppressed, add it to the list
				report.append(err)

		assert_no_instrument_status_errors(self._io.resource_name, report, "InstrErrorSuppressor context -")
		return suppressed


class VisaTimeoutSuppressor:
	"""Context-manager class to suppress exception caused by the VISA Timeout.
	Careful!!!: The VisaTimeoutSuppressor only suppresses the very first VISA Timeout exception,
	and afterward it ends. Therefore, use only one command per context manager,
	if you do not want to skip the following ones.
	:param io: RsInstrument session instance.
	:param visa_tout_ms: VISA Timeout in milliseconds, that is set for this context.
	"""
	def __init__(self, io: Instrument, visa_tout_ms: int = 0):
		self._io: Instrument = io
		self._old_visa_tout_ms: int = 0

		self._timeout_occurred: bool = False

		if visa_tout_ms > 0:
			self._old_visa_tout_ms = io.visa_timeout
			io.visa_timeout = visa_tout_ms

	def __enter__(self) -> 'VisaTimeoutSuppressor':
		"""Stuff to do when entering the context.
		Currently, do not do anything."""
		return self

	def __exit__(self, exc_type, value, traceback):
		"""Stuff to do when leaving the context:
				- Sets the VISA Timeout to the original value.
				- Sets the query_instr_status to the original value.
			:returns: True, if the exception in the block should be suppressed."""
		if self._old_visa_tout_ms > 0:
			self._io.visa_timeout = self._old_visa_tout_ms

		# Was it straight timeout exception?
		if isinstance(value, TimeoutException):
			self._timeout_occurred = True
			if not self._io.query_instr_status:
				# If the instrument status checking is OFF, clear the error queue
				try:
					self._instrument_errors = self._io.query_all_syst_errors(False)
				except VisaIOError:
					pass

			return True

		# Instrument status exception, but caused primarily by timeout exception
		if isinstance(value, StatusException):
			if value.first_exc is not None and value.first_exc is TimeoutException:
				self._timeout_occurred = True
				if not self._io.query_instr_status:
					# If the instrument status checking is OFF, clear the error queue
					try:
						self._instrument_errors = self._io.query_all_syst_errors(False)
					except VisaIOError:
						pass

				return True

		if isinstance(value, VisaIOError) and hasattr(value, 'source'):
			# check for the source _narrow_down_tout_error
			if value.source == '_narrow_down_io_tout_error':
				self._timeout_occurred = True
				return True

		return False

	def get_timeout_occurred(self) -> bool:
		"""Returns True, if the VISA timeout occurred in the context."""
		return self._timeout_occurred



from typing import Tuple, Collection
from ..Internal.Core import Core


class ErrorQueryResult:
    """Structure class for holding one error query result."""

    def __init__(self, code: int, message: str) -> None:
        self._code = code
        self._message = message

    @property
    def code(self) -> int:
        return self._code

    @property
    def message(self) -> str:
        return self._message


class IviUtility:
    """Utilities class required by the IVI-Python standard."""

    def __init__(self, core: Core):
        self._core = core

    @property
    def driver_version(self) -> str:
        """Returns the instrument driver version."""
        return self._core.driver_version

    @property
    def driver_vendor(self) -> str:
        """Returns the driver manufacturer: 'ROHDE&SCHWARZ'."""
        return "ROHDE&SCHWARZ"

    @property
    def instrument_manufacturer(self) -> str:
        """Returns the manufacturer of the instrument."""
        return self._core.io.manufacturer

    @property
    def instrument_model(self) -> str:
        """Returns the current instrument's full name e.g. 'FSW26'."""
        return self._core.io.full_model_name

    @property
    def query_instrument_status_enabled(self) -> bool:
        """Sets / returns Instrument Status Checking.
        When True (default is True), all the driver methods and properties are sending "SYSTem:ERRor?"
        at the end to immediately react on error that might have occurred.
        We recommend to keep the state checking ON all the time. Switch it OFF only in rare cases when you require maximum speed.
        The default state after initializing the session is ON."""
        return self._core.io.query_instr_status

    @query_instrument_status_enabled.setter
    def query_instrument_status_enabled(self, value) -> None:
        """Sets / returns Instrument Status Checking.
        When True (default is True), all the driver methods and properties are sending "SYSTem:ERRor?"
        at the end to immediately react on error that might have occurred.
        We recommend to keep the state checking ON all the time. Switch it OFF only in rare cases when you require maximum speed.
        The default state after initializing the session is ON."""
        self._core.io.query_instr_status = value

    @property
    def simulation_enabled(self) -> bool:
        """Read-only property indicating if simulation mode is enabled. This property is only settable in the driver's constructor options parameter 'Simulate'."""
        return self._core.simulating

    @property
    def supported_instrument_models(self) -> Tuple[str, ...]:
        """Returns models supported by the driver, one per element."""
        return self._core.supported_instr_models

    def reset(self) -> None:
        """Resets the instrument and clears its status.
        This is typically done by sending the *RST and *CLS SCPI Commands."""
        self._core.io.reset()

    def error_query(self) -> ErrorQueryResult | None:
        """Returns the last error in the instrument's error queue.
        Returns None if no error is present."""
        err = self._core.io.query_syst_error(include_code=True, enable_log=True)
        if err is None:
            return None
        return ErrorQueryResult(err[0], err[1])

    def error_query_all(self) -> Collection[ErrorQueryResult]:
        """Returns all the errors currently reported in the instrument's error queue.
          If no error is present, the method returns an empty collection."""
        errs = self._core.io.query_all_syst_errors(include_codes=True, enable_log=True)
        if errs is None:
            return []
        ret_val = [ErrorQueryResult(errs[x][0], errs[x][1]) for x in range(len(errs))]
        return ret_val

    def raise_on_device_error(self) -> None:
        """Calls error_query_all() and raises an exception if any instrument errors were detected."""
        self._core.io.check_status_always(log_ok_result=True)






"""Class for R&S Mobile Radio Test instruments that use reliability indicators."""


from time import localtime, strftime
from typing import Callable, Any

from ..Internal.ArgLinkedEventArgs import ArgLinkedEventArgs
from ..Internal.Core import Core

codes_table = {
				0:       'OK',
				1:       'Measurement Timeout',
				2:       'Capture Buffer Overflow',
				3:       'Over-driven',
				4:       'Under-driven',
				6:       'Trigger Timeout',
				7:       'Acquisition Error',
				8:       'Sync Error',
				9:       'Uncalibrated',
				15:      'Reference Frequency Error',
				16:      'RF Not Available',
				17:      'RF Level not Settled',
				18:      'RF Frequency not Settled',
				19:      'Call not Established',
				20:      'Call Type not Usable',
				21:      'Call Lost',
				23:      'Missing Option',
				24:      'Invalid RF Setting',
				26:      'Resource Conflict',
				27:      'No Sensor Connected',
				28:      'Unexpected Parameter Change',
				30:      'File not Found',
				31:      'No DTM reply',
				32:      'ACL Disconnected',
				40:      'ARB File CRC Error',
				42:      'ARB Header Tag Invalid',
				43:      'ARB Segment Overflow',
				44:      'ARB File not Found',
				45:      'ARB Memory Overflow',
				46:      'ARB Sample Rate out of Range',
				47:      'ARB Cycles out of Range',
				50:      'Startup Error',
				51:      'No Reply',
				52:      'Connection Error',
				53:      'Configuration Error',
				54:      'Filesystem Error',
				60:      'Invalid RF-Connector Setting',
				93:      'OCXO Oven Temperature too low',
				101:     'Firmware Error',
				102:     'Unidentified Error',
				103:     'Parameter Error',
				104:     'Not Functional'}


class ReliabilityEventArgs:
	"""Arguments for reliability indicator event."""

	def __init__(self, timestamp, code: int, message: str, context: str):
		self.timestamp = timestamp
		self.message = message
		self.code = code
		self.message = message
		self.context = context


class Reliability:
	"""Reliability class that handles all the necessary tasks related to reliability indicator."""

	def __init__(self, core: Core):
		self._core: Core = core
		self._last_value: int = 0
		self._last_context: str = ''
		self._last_timestamp = None
		self._exception_on_error = False
		# noinspection PyTypeChecker
		self._on_update_handler: Callable = None
		self._core.set_link_handler('Reliability', self._permanent_on_update_handler)

	@property
	def last_value(self) -> int:
		"""Returns the last updated Reliability code."""
		return self._last_value

	@property
	def last_context(self) -> str:
		"""Returns the last updated Context of the reliability code - usually the SCPI query on which the instrument responded with the Reliability code."""
		return self._last_context

	@property
	def last_timestamp(self) -> Any:
		"""Returns the time of the last Reliability update."""
		return self._last_timestamp

	@property
	def last_message(self) -> str:
		"""Returns the LastValue of the reliability table converted to human-readable string."""
		if self._last_value in codes_table:
			return codes_table[self._last_value]
		else:
			return f'Undefined reliability code {self._last_value}.'

	@property
	def exception_on_error(self) -> bool:
		"""If True, (default is False) the object throws an exception if the updated reliability is not 0 (non-OK)."""
		return self._exception_on_error

	@exception_on_error.setter
	def exception_on_error(self, value) -> None:
		"""If True, (default is False) the object throws an exception if the updated reliability is not 0 (non-OK)."""
		self._exception_on_error = value

	def on_update_handler(self, handler: Callable) -> None:
		"""Register the handler for on_update event.
		This handler is invoked with each update of the reliability indicator.
		Handler API: handler(event_args: ReliabilityEventArgs)"""
		self._on_update_handler = handler

	def _permanent_on_update_handler(self, event_args: ArgLinkedEventArgs) -> None:
		"""Permanent on_update handler. Takes care of updating all the 'last_xxx' values and calling a user-defined updated_handler."""
		self._last_value = int(str(event_args.value))
		self._last_context = event_args.context
		self._last_timestamp = event_args.timestamp
		if self._on_update_handler:
			# Call the additional handler if registered
			rel_events_args = ReliabilityEventArgs(self._last_timestamp, self._last_value, self.last_message, self._last_context)
			self._on_update_handler(rel_events_args)
		if self._exception_on_error and self._last_value != 0:
			raise Exception(
				f'Reliability indicator error. Time: {strftime("%H:%M:%S", localtime(self._last_timestamp))}, '
				f'Context: {self._last_context}, Value {self._last_value}: {self.last_message}')

	def sync_from(self, source: 'Reliability') -> None:
		"""Synchronizes this Reliability with the source."""
		self.exception_on_error = source.exception_on_error
		self.on_update_handler(source._on_update_handler)



"""Class defining Linked argument Event."""


class ArgLinkedEventArgs(object):
	"""Contains event data for suppressed argument."""

	def __init__(self, link_name: str, arg_name: str, value: object = None, context: str = '', timestamp: float | None = None):
		self.link_name = link_name
		self.arg_name = arg_name
		self.value = value
		self.context = context
		self.timestamp = timestamp

	def __str__(self):
		result = f"ArgLinkedEventArgs '{self.link_name}'"
		if self.arg_name:
			result += f"argument name '{self.arg_name}'"
		result += f"value: {self.value}"
		return result



"""Visa Session is an extension of the pure VISA providing higher level of methods regardless of the session kind."""

import time
from enum import Enum, Flag
from typing import List, Tuple, Callable, AnyStr
import os.path
import re
import threading

# noinspection PyPackageRequirements
import pyvisa
from pyvisa.errors import StatusCode, VisaIOError

from .VisaPluginSocketIo import ResourceManager, SocketIo
from . import Conversions as Conv
from .InstrumentErrors import ResourceError, RsInstrException, TimeoutException, StatusException
from .InstrumentErrors import throw_opc_tout_exception, throw_bin_block_unexp_resp_exception, assert_no_instrument_status_errors
from .InstrumentErrors import assert_query_has_qmark, assert_cmd_has_no_qmark

from .InstrumentSettings import InstrumentSettings, WaitForOpcMode, OpcSyncQueryMechanism, InstrViClearMode as ViClearMode
from .StreamReader import StreamReader
from .StreamWriter import StreamWriter
from .Utilities import size_to_kb_mb_string, calculate_chunks_count
import platform
import struct


class SessionKind(Enum):
	"""Visa instrument session type."""
	unsupported = 0
	gpib = 1
	serial = 2
	vxi11 = 3
	socket = 5
	usb = 6
	rs_nrp = 7


class ReadDataType(Enum):
	"""Data type returned by the instrument."""
	unknown = 0
	ascii = 1
	null = 2
	bin_known_len = 3
	bin_unknown_len = 4


class StatusByte(Flag):
	"""Status Byte flags."""
	NONE = 0x00
	custom_bit_0 = 0x01
	custom_bit_1 = 0x02
	error_queue_not_empty = 0x04
	questionable_status_reg = 0x08
	message_available = 0x10
	event_status_byte = 0x20
	request_service = 0x40
	operation_status_reg = 0x80


class EventStatusRegister(Flag):
	"""Event Status Register flags."""
	null = 0x00
	operation_complete = 0x01


# noinspection PyUnresolvedReferences
class VisaSession(object):
	"""Extended VISA class."""

	def __init__(self, resource_name: str, settings: InstrumentSettings, direct_session=None):
		self.reusing_session = direct_session is not None
		# noinspection PyTypeChecker
		self._data_chunk_size: int = None
		self._std_bin_block_header_max_len: int = 999999999
		self._lock = None
		self._flush_with_tout_tolerance: None | bool = None
		self.disable_opc_query: bool = settings.disable_opc_query
		self.last_status = None
		self.visa_library_name = None
		self.resource_name = resource_name  # might be changed later if direct_session is used
		self.encoding = settings.encoding  # default encoder between bytes and string
		self.cmd_idn = settings.cmd_idn
		self.skip_status_system_setting = settings.skip_status_system_setting
		self.skip_clear_status = settings.skip_clear_status
		self.stb_in_error_check = settings.stb_in_error_check
		self.opc_sync_query_mechanism = settings.opc_query_sync_mechanism
		self.each_cmd_prefix = settings.each_cmd_prefix
		self.each_cmd_suffix = settings.each_cmd_suffix
		self.strip_str_trailing_ws = settings.strip_str_trailing_ws

		# Implemented for interface compatibility with VisaSessionSim
		self.cached_to_stream = False

		# Event handlers
		# noinspection PyTypeChecker
		self.on_read_chunk_handler: Callable = None
		"""If assigned a handler, the VisaSession sends it event on each read chunk transfer."""
		# noinspection PyTypeChecker
		self.on_write_chunk_handler: Callable = None
		"""If assigned a handler, the VisaSession sends it event on each write chunk transfer."""
		self.io_events_include_data: bool = False
		"""If true, the VisaSession events sent to on_read_chunk_handler and on_write_chunk_handler contain transferred data."""

		if self.reusing_session:
			# Reuse the session
			self._session = VisaSession.get_and_check_direct_session(direct_session)
			self.resource_name = self._session.resource_name
		else:
			# Create new session
			# Check resource_name for the trailing (SelectVisa=..)
			pure_resource_name, visa_select = self._get_pure_resource_name(resource_name)
			if settings.visa_select is not None:
				visa_select = settings.visa_select
			self._rm = VisaSession.get_resource_manager(visa_select)
			self.manufacturer = self._get_visa_manufacturer()

			# Resource manager opening
			try:
				acc_mode = pyvisa.constants.AccessModes.no_lock if settings.exclusive_lock is False else pyvisa.constants.AccessModes.exclusive_lock
				self._session = self._rm.open_resource(resource_name=pure_resource_name, open_timeout=settings.open_timeout, access_mode=acc_mode)
			except pyvisa.VisaIOError as e:
				if e.error_code != pyvisa.constants.StatusCode.error_resource_not_found:
					raise e
				message = e.description
				message += f"\nLibrary: {self._rm.visalib}\nManufacturer: {self.manufacturer}\nResource Name: '{resource_name}'"
				raise ResourceError(resource_name, message)
			self.resource_name = resource_name

		# Decide, whether to create a new thread lock or the existing one from the session
		if hasattr(self._session, 'session_thread_rlock'):
			rlock = self._session.session_thread_rlock
			if isinstance(rlock, type(threading.RLock())):
				self.assign_lock(rlock)
		if self.get_lock() is None:
			# The existing session did not have a thread lock, assign a new one
			self.assign_lock(threading.RLock())

		self._interface_type = SessionKind.unsupported
		if self._session.interface_type == pyvisa.constants.InterfaceType.gpib or self._session.interface_type == pyvisa.constants.InterfaceType.gpib:
			self._interface_type = SessionKind.gpib

		elif self._session.interface_type == pyvisa.constants.InterfaceType.rsnrp:
			self._interface_type = SessionKind.rs_nrp

		elif self._session.interface_type == pyvisa.constants.InterfaceType.asrl:
			self._interface_type = SessionKind.serial

		elif self._session.interface_type == pyvisa.constants.InterfaceType.usb:
			# Check whether it is not the NRP-Z
			intf_type = self._session.get_visa_attribute(pyvisa.constants.VI_ATTR_INTF_TYPE)
			if intf_type == pyvisa.constants.InterfaceType.rsnrp:
				self._interface_type = SessionKind.rs_nrp

		elif self._session.interface_type == pyvisa.constants.InterfaceType.tcpip:
			self._interface_type = SessionKind.vxi11
			if self._session.resource_class == 'SOCKET':
				self._interface_type = SessionKind.socket

		# Specifics for different interfaces
		self._assure_write_with_tc = settings.assure_write_with_tc
		self._term_char = settings.term_char
		self._term_char_bin = self._term_char.encode(self.encoding)
		self._session.write_termination = ''
		self.vxi_capable = settings.vxi_capable

		if self._interface_type == SessionKind.serial:
			self.vxi_capable = False
		elif self._interface_type == SessionKind.socket:
			self.vxi_capable = False

		# NRP-Z specific settings
		if self.is_rsnrp_session():
			self.disable_opc_query = True
			# NRP-Z does not support chunk reading, therefore the segment must be in one piece
			settings.io_segment_size = 1000000
			self.vxi_capable = False

		self.write_delay = settings.write_delay
		self.read_delay = settings.read_delay
		self._viclear_exe_mode = settings.viclear_exe_mode
		self._opc_wait_mode = settings.opc_wait_mode

		# Parameters that need to be coerced based on Vxi-capability
		if self.vxi_capable:
			self._add_term_char_to_write_bin_block = settings.add_term_char_to_write_bin_block
			self._session.read_termination = ''
		else:
			self._add_term_char_to_write_bin_block = True
			self._session.read_termination = self._term_char
			self._assure_write_with_tc = True

		# Changeable settings
		self.opc_timeout = 10000 if settings.opc_timeout == 0 else settings.opc_timeout
		self.visa_timeout = settings.visa_timeout
		self._session.chunk_size = settings.io_segment_size
		self._data_chunk_size = settings.io_segment_size

		# Must call the VISA viClear() before any communication with the instrument
		self.clear()

		# Further steps are for NRP-Z session not valid
		if self.is_rsnrp_session():
			return

		# First commands, can be more than one, separated by ';;'
		if settings.first_cmds:
			cmds = settings.first_cmds.split(';;')
			for cmd in cmds:
				if cmd.startswith('<w>'):
					self.write(cmd[3:])
				elif cmd.startswith('<q>'):
					_ = self._query_str_no_events(cmd[3:])
				elif '?' in cmd:
					_ = self._query_str_no_events(cmd)
				else:
					self.write(cmd)

		# Clear instrument status
		if not self.skip_clear_status:
			self.write('*CLS')
			if self.vxi_capable:
				try:
					stb = self._read_stb()
				except VisaIOError as err:
					msg = 'Operation Read STB: ' + err.description
					raise RsInstrException(msg)
				if stb & StatusByte.message_available:
					self._flush_junk_data()

		# Apply settings for ESE and SRE, plus coerce the _opcWaitMode if necessary
		self._opc_wait_mode = self._set_regs_ese_sre(self._opc_wait_mode)

	@staticmethod
	def get_and_check_direct_session(direct_session):
		"""Returns direct session if it's a proper type.
		If the direct_session is None, the function returns None.
		If the direct_session is of an unsupported type, the function raises RsInstrException."""
		if direct_session is None:
			return None
		# Reuse the session
		if not isinstance(direct_session, pyvisa.Resource) and not isinstance(direct_session, SocketIo):
			raise RsInstrException(f"Direct_session must be a VISA resource object. Actual type: '{type(direct_session)}', value: '{direct_session}'")
		return direct_session

	@staticmethod
	def _get_pure_resource_name(resource_name: str):
		"""Returns pure resource name stripped of the (SelectVisa) part and the visa_select string"""
		m = re.search(r'(.+)\(SelectVisa=([^),]+)\)', resource_name)
		if not m:
			return resource_name, None
		resource_name = m.group(1).strip()
		visa_select = m.group(2).strip()
		return resource_name, visa_select

	# noinspection SpellCheckingInspection
	@classmethod
	def get_resource_manager(cls, visa_select: str) -> pyvisa.ResourceManager:
		"""Returns resource manager for the desired VISA implementation"""
		operating_system = platform.system().lower()
		vsl = None if visa_select is None else visa_select.lower()
		bittness = struct.calcsize('P') * 8
		# Try if you find the default VISA dll
		try:
			if visa_select is None or visa_select in ['@default', '@standard', 'default', 'standard', 'defaultvisa', 'standardvisa', '@defaultvisa', '@standardvisa']:
				return pyvisa.ResourceManager()

			if vsl in ['@ni', 'ni', 'ivi', '@ivi', 'visa-ni', 'nivisa', 'ni-visa', 'nationalinstruments', 'nationalinstrumentsvisa']:
				return pyvisa.ResourceManager()

			if vsl in ['@py', 'pyvisa', 'visa-py', 'pyvisa-py']:
				return pyvisa.ResourceManager('@py')
		except ValueError:
			# None of the required implementations found, fall back to the R&S VISA
			visa_select = 'rsvisa'
			vsl = visa_select.lower()

		# from here, RsVisa implementation is considered
		if 'rohde&schwarz' in vsl or 'rohdeschwarz' in vsl or vsl == 'rsvisa' or vsl == 'rs' or vsl == 'r&s':
			if operating_system == 'windows':
				if bittness == 32:
					visa_select = r'c:\Windows\SysWOW64\RsVisa32.dll'
				else:
					visa_select = r'c:\Windows\system32\RsVisa32.dll'
				return pyvisa.ResourceManager(visa_select)
			elif operating_system == 'linux':
				# The default install location may be different
				# for debian/red hat/opensuse derived distributions
				check_visa = [f'/usr/lib{bittness}/librsvisa.so', r'/usr/lib/librsvisa.so']
				for check in check_visa:
					if os.path.isfile(check):
						return pyvisa.ResourceManager(check)
			elif operating_system == 'darwin':
				# MacOS
				check_visa = [f'/Library/Frameworks/RsVisa.framework/Versions/Current/RsVisa/librsvisa.dylib']
				for check in check_visa:
					if os.path.isfile(check):
						return pyvisa.ResourceManager(check)

		if vsl in ['socketio', 'socket', 'none']:
			return ResourceManager()

		return pyvisa.ResourceManager(visa_select)

	def _get_visa_manufacturer(self) -> str:
		"""Returns manufacturer of the current VISA"""
		if hasattr(self._rm, 'VisaManufacturerName'):
			return self._rm.VisaManufacturerName
		try:
			return self._rm.visalib.get_attribute(self._rm.session, pyvisa.constants.VI_ATTR_RSRC_MANF_NAME)[0]
		except TypeError:
			return self._rm.visalib.__class__.__name__

	def is_rsnrp_session(self) -> bool:
		"""Returns True, if the current session is a NRP-Z session"""
		return self._interface_type == SessionKind.rs_nrp

	def assign_lock(self, lock: threading.RLock) -> None:
		"""Assigns the provided thread lock by setting the pyvisa runtime session attribute 'session_thread_rlock'.
		This is done, because if the session is to be entered as an existing session to another RsInstrument object,
		the lock must be shared as well. The lock is only used by the parent class Instrument."""
		setattr(self._session, 'session_thread_rlock', lock)
		self._lock = lock

	def get_lock(self) -> threading.RLock:
		"""Returns the current RLock object."""
		return self._lock

	def lock_resource(self, timeout: int, requested_key: str | bytes = None) -> bytes | str | None:
		"""Locks the instrument to prevent it from communicating with other clients."""
		if requested_key is None:
			self._session.lock_excl(timeout)
			return None
		else:
			return self._session.lock(timeout, requested_key)

	def unlock_resource(self) -> None:
		"""Unlocks the instrument to other clients."""
		self._session.unlock()

	@property
	def visa_timeout(self) -> int:
		"""See the visa_timeout.setter."""
		return int(self._session.timeout)

	@visa_timeout.setter
	def visa_timeout(self, value: int) -> None:
		"""Sets / Gets visa IO timeout in milliseconds."""
		self._session.timeout = int(value)

	@property
	def data_chunk_size(self) -> int:
		"""Returns max chunk size of one data block."""
		return self._data_chunk_size

	@data_chunk_size.setter
	def data_chunk_size(self, chunk_size: int) -> None:
		"""Sets the maximum size of one block transferred during write/read operations."""
		self._data_chunk_size = int(chunk_size)
		self._session.chunk_size = int(chunk_size)

	def _resolve_opc_timeout(self, timeout: int) -> int:
		"""Resolves entered timeout value - if the input value is less than 1, it is replaced with opc_timeout."""
		if timeout is None or timeout < 1:
			return self.opc_timeout
		else:
			return timeout

	def _set_regs_ese_sre(self, mode: WaitForOpcMode) -> WaitForOpcMode:
		"""Based on the WaitForOpcMode, it sets the ESE and SRE register masks.
		Returns coerced WaitForOpcMode."""
		# Set the SRE and ESE registers accordingly
		# No SRE is supported
		if self.skip_status_system_setting:
			return mode
		self._set_ese_mask(EventStatusRegister.operation_complete)
		self._set_sre_mask(StatusByte.NONE)
		return mode

	# noinspection PyTypeChecker
	def _set_ese_mask(self, mask: EventStatusRegister, reset: bool = True) -> None:
		"""Sends *ESE command with mask parameter."""
		if not reset:
			current_value = int(self._query_str_no_events('*ESE?'))
			mask = current_value | mask.value
		self.write("*ESE %d" % mask.value)

	# noinspection PyTypeChecker
	def _set_sre_mask(self, mask: StatusByte, reset: bool = True) -> None:
		"""Sends *SRE command with StatusByte mask parameter."""
		if not reset:
			current_value = int(self._query_str_no_events('*SRE?'))
			mask = current_value | mask.value
		# Also affect the _opc_wait_mode:
		# If the mask has event_status_byte == false, and the _opc_wait_mode is service_request, set it to stb_poll
		# If the mask has event_status_byte == true, do not change anything
		self.write(f'*SRE {mask.value}')

	def _write_and_poll_stb_vxi(self, command: str, is_query: bool, timeout: int) -> StatusByte:
		"""Reads Status Byte Register and ends if the ESB bit (5) is set to 1.
		Also works with the SOCKET and SERIAL interface by sending *STB? query.
		In that case however, command cannot be a query.
		Returns the last read Status Byte value."""
		timeout_secs = timeout / 1000
		# noinspection PyTypeChecker
		end_mask = StatusByte.error_queue_not_empty | StatusByte.event_status_byte
		if command.endswith(self._term_char):
			command = command.rstrip(self._term_char)

		if is_query:
			if self.opc_sync_query_mechanism == OpcSyncQueryMechanism.standard or self.opc_sync_query_mechanism == OpcSyncQueryMechanism.also_check_mav:
				self.clear_before_read()
				self.write(command + ';*OPC')
				if self.opc_sync_query_mechanism == OpcSyncQueryMechanism.also_check_mav:
					end_mask |= StatusByte.message_available

			elif self.opc_sync_query_mechanism == OpcSyncQueryMechanism.only_check_mav_err_queue:
				self.write(command)
				# noinspection PyTypeChecker
				end_mask = StatusByte.error_queue_not_empty | StatusByte.message_available

			elif self.opc_sync_query_mechanism == OpcSyncQueryMechanism.cls_only_check_mav_err_queue:
				self.clear_before_read()
				self.write(command)
				# noinspection PyTypeChecker
				end_mask = StatusByte.error_queue_not_empty | StatusByte.message_available
		else:
			self.clear_before_read()
			self.write(command + ';*OPC')

		start = time.time()
		# STB polling loop
		while True:
			stb = self._read_stb()
			elapsed = self._polling_delay(start)
			if elapsed > timeout_secs:
				self._narrow_down_opc_tout_error(command, is_query, timeout)
			if end_mask & stb:
				break
		return stb

	def _write_and_poll_stb_non_vxi(self, command: str, timeout: int) -> StatusByte:
		"""Queries Status Byte Register (*STB?) and ends if the ESB bit (5) is set to 1.
			The command must not be a query. Also works with the SOCKET and SERIAL interface.
			Returns the last read Status Byte value."""
		timeout_secs = timeout / 1000
		# noinspection PyTypeChecker
		end_mask = StatusByte.error_queue_not_empty | StatusByte.event_status_byte
		self.clear_before_read()
		if command.endswith(self._term_char):
			command = command.rstrip(self._term_char)
		self.write(command + ';*OPC')
		start = time.time()
		# STB polling loop
		while True:
			stb = self._query_stb()
			elapsed = self._polling_delay(start)
			if elapsed > timeout_secs:
				throw_opc_tout_exception(self.opc_timeout, timeout)
			if stb & end_mask:
				break
		return stb

	def _narrow_down_opc_tout_error(self, command: str, is_query: bool, timeout: int) -> None:
		"""Called by the _write_and_poll_stb_vxi when the timeout expires.
		The method tries to closer identify the cause of the timeout."""
		stb = self._read_stb()
		timeout = self._resolve_opc_timeout(timeout)
		if is_query:
			if stb & StatusByte.error_queue_not_empty:
				self.clear()
				context = f"Query '{command.strip()}' with OPC Wait resulted in timeout. OPC Timeout is set to {timeout} ms. Additionally, "
				assert_no_instrument_status_errors(self.resource_name, self.query_all_syst_errors(), context, first_exc=TimeoutException)
			throw_opc_tout_exception(self.opc_timeout, timeout, f"Query '{command.strip()}'.")
		else:
			if stb & StatusByte.error_queue_not_empty:
				self.clear()
				context = f"Command '{command.strip()}' with OPC Wait resulted in timeout. OPC Timeout is set to {timeout} ms. Additionally, "
				assert_no_instrument_status_errors(self.resource_name, self.query_all_syst_errors(), context, first_exc=TimeoutException)
			throw_opc_tout_exception(self.opc_timeout, timeout, f"Command '{command.strip()}'.")

	def _narrow_down_io_tout_error(self, context: str, visa_timeout: int = 0) -> None:
		"""Called internally after IOTimeoutException can narrow down the error to more specific exception.
		You can define the visa_timeout value for the error message. Otherwise, the current visa_timeout is reported."""
		context_stripped = context.strip().rstrip("- ")
		if not self.stb_in_error_check:
			raise TimeoutException(context_stripped)

		try:
			if self.vxi_capable:
				stb = self._read_stb()
			elif self.is_rsnrp_session():
				stb = StatusByte(StatusByte.NONE)
			else:
				# Non-Vxi session
				old_tout = self.visa_timeout
				try:
					self.visa_timeout = 500
					stb = self._query_stb(False)
				finally:
					self.visa_timeout = old_tout
			if visa_timeout <= 0:
				visa_timeout = self.visa_timeout

			context = context + f'VISA Timeout error occurred ({visa_timeout} milliseconds)'
			if stb & StatusByte.error_queue_not_empty:
				assert_no_instrument_status_errors(self.resource_name, self.query_all_syst_errors(), context + ' and ...', first_exc=TimeoutException)
			# In case none of the previous exceptions is thrown
			raise TimeoutException(context)

		except pyvisa.VisaIOError as e:
			e.source = '_narrow_down_io_tout_error'
			raise e

	def _polling_delay(self, start):
		"""Generates progressive polling delay."""

		elapsed = time.time() - start
		if self._opc_wait_mode == WaitForOpcMode.stb_poll:
			if elapsed < 0.01:
				return elapsed
			if elapsed < 0.1:
				time.sleep(0.005)
				return elapsed
			if elapsed < 1:
				time.sleep(0.02)
				return elapsed
			if elapsed < 5:
				time.sleep(0.05)
				return elapsed
			if elapsed < 10:
				time.sleep(0.1)
				return elapsed
			if elapsed < 50:
				time.sleep(0.5)
				return elapsed
			time.sleep(1)
		elif self._opc_wait_mode == WaitForOpcMode.stb_poll_slow:
			if elapsed < 0.01:
				time.sleep(0.001)
				return elapsed
			if elapsed < 1:
				time.sleep(0.02)
				return elapsed
			if elapsed < 5:
				time.sleep(0.1)
				return elapsed
			if elapsed < 10:
				time.sleep(0.2)
				return elapsed
			if elapsed < 20:
				time.sleep(0.5)
				return elapsed
			time.sleep(1)
		elif self._opc_wait_mode == WaitForOpcMode.stb_poll_superslow:
			if elapsed < 1:
				time.sleep(0.1)
				return elapsed
			if elapsed < 10:
				time.sleep(0.5)
				return elapsed
			if elapsed < 20:
				time.sleep(1)
				return elapsed
			time.sleep(2)

		return elapsed

	@staticmethod
	def _parse_err_query_response(response: str) -> Tuple[int, str]:
		"""
		Parses entered response string to Tuple(code, message).
		E.g.: response = '-110,"Command error"' returns: (-110,'Command error')
		"""
		m = re.match(r'([-+]?\d+).*?[\'"](.*)[\'"]', response)
		code = 0
		if m:
			try:
				code = int(m.group(1))
			except ValueError:
				pass
			return code, m.group(2)
		else:
			return code, response

	def query_syst_error(self) -> Tuple[int, str] | None:
		"""Returns one response to the SYSTEM:ERROR? query.
		The response is a Tuple of (code: int, message: str)"""
		error = self._query_str_no_events('SYST:ERR?')
		if error.startswith('0,') or error.startswith('+0,'):
			return None
		return self._parse_err_query_response(error.strip())

	def query_all_syst_errors(self) -> List[Tuple[int, str]] | None:
		"""Returns all errors in the instrument's error queue.
		If no error is detected, the return value is None."""
		errors = []
		while True:
			entry = self.query_syst_error()
			if entry is None:
				break
			errors.append(entry)
			if len(errors) > 50:
				# Safety stop
				errors.append('query_all_syst_errors - max limit 50 of SYST:ERR? sent.')
				break
		if len(errors) == 0:
			return None
		else:
			return errors

	def _query_stb(self, allow_tout_error_narrow_down: bool = True) -> StatusByte:
		"""Sends *STB? query and reads the result."""
		return StatusByte(int(self._query_str_no_events('*STB?', allow_tout_error_narrow_down)))

	def _read_stb(self) -> StatusByte:
		"""Calls viReadStb and returns the result."""
		return StatusByte(self._session.read_stb())

	def clear_before_read(self) -> None:
		"""Clears IO buffers and the ESR register before reading/writing responses synchronized with *OPC."""

		# For NRP-Z sessions, skip this completely
		if self.is_rsnrp_session() or self.skip_clear_status:
			return

		if not self.vxi_capable:
			# Non-Vxi session must use *CLS in any case
			self.write('*CLS')
			correct = False
			if self.disable_opc_query:
				opc = '1'
			else:
				opc = self._query_str_no_events('*OPC?')

			repeat = 0
			while not correct:
				if opc is None:
					break

				if len(opc) <= 2:
					opc = opc.strip()
					correct = opc == '0' or opc == '1'
				if not correct:
					# Read again with a small VISA timeout
					opc = self._read_str_timed(5, True)
				repeat += 1
				if repeat > 10:
					break

		stb = self._query_stb()
		# noinspection PyTypeChecker
		condition = StatusByte.error_queue_not_empty | StatusByte.message_available | StatusByte.event_status_byte
		if not stb & condition:
			return
		repeat = 0
		# Loop more times to clear the status subsystem
		while stb & condition:
			if stb & StatusByte.error_queue_not_empty:
				self.write('*CLS')
				_ = self.query_all_syst_errors()
			if stb & StatusByte.message_available:
				# Clear output buffer
				self._flush_junk_data()
			if stb & StatusByte.event_status_byte:
				# OPC or error bits in the ESR
				self.write('*CLS')
				self.query_and_clear_esr()
			# Check if the status byte value changed
			previous_stb = stb
			stb = self._query_stb()
			if stb == previous_stb:
				repeat += 1
				if repeat > 10:
					raise RsInstrException(f"Cannot clear the instrument's status subsystem. Status Byte: '{stb}'")

	def _flush_junk_data(self) -> None:
		"""Reads junk bytes to clear the instrument's output buffer."""
		if self.read_delay > 0:
			time.sleep(self.read_delay / 1000)

		if self._flush_with_tout_tolerance is None:
			res = self._read_str_timed(300, True)
			self._flush_with_tout_tolerance = True if res is None else False
		elif self._flush_with_tout_tolerance:
			self._read_str_timed(300, True)
		else:
			self._read_unknown_len(StreamWriter.as_bin_var(), False)

	def clear(self) -> None:
		"""Perform VISA viClear conditionally based on the instrument settings."""
		perform_all = ViClearMode.execute_on_all in self._viclear_exe_mode
		perform = False
		if perform_all:
			perform = True
		else:
			# Perform on all is blocked, use the SessionKind to decide
			if self._interface_type == SessionKind.gpib:
				perform = ViClearMode.execute_on_gpib in self._viclear_exe_mode

			elif self._interface_type == SessionKind.serial:
				perform = ViClearMode.execute_on_serial in self._viclear_exe_mode

			elif self._interface_type == SessionKind.socket:
				perform = ViClearMode.execute_on_socket in self._viclear_exe_mode

			elif self._interface_type == SessionKind.usb:
				perform = ViClearMode.execute_on_usb in self._viclear_exe_mode

			elif self._interface_type == SessionKind.vxi11:
				perform = ViClearMode.execute_on_tcpvxi in self._viclear_exe_mode

		if not perform:
			return

		if ViClearMode.ignore_error in self._viclear_exe_mode:
			# noinspection PyBroadException
			try:
				self._session.clear()
			except Exception:
				pass
		else:
			self._session.clear()

	def is_connection_active(self) -> bool:
		"""Returns true, if the VISA connection is active and the communication with the instrument still works.
		This is achieved by:
		- checking the session property timeout
		- sending the *IDN? query"""
		if self._session is None:
			return False
		# noinspection PyBroadException
		try:
			old_tout = self.visa_timeout
			self.visa_timeout = 2000
			if len(self.cmd_idn) > 0:
				self.write(self.cmd_idn)
				_ = self._read_str_no_events()
			self.visa_timeout = old_tout
			return True
		except Exception:
			return False

	def _write_and_wait_for_opc(self, command: str, is_query: bool, timeout: int) -> StatusByte:
		"""Internal method to synchronize a command with OPC timeout.
		Timeout value 0 means the OPC timeout is used."""
		timeout = self._resolve_opc_timeout(timeout)

		if command.endswith(self._term_char):
			command = command.rstrip(self._term_char)
		if is_query:
			assert_query_has_qmark(command, 'Query with OPC')
		else:
			assert_cmd_has_no_qmark(command, 'Write with OPC')

		if self._opc_wait_mode == WaitForOpcMode.opc_query:
			if is_query:
				raise RsInstrException('Sending a query with OpcQuery synchronization is not possible')
			stb = self._write_and_query_opc(command, timeout)
		else:
			# STB polling
			if self.vxi_capable:
				stb = self._write_and_poll_stb_vxi(command, is_query, timeout)
			else:
				stb = self._write_and_poll_stb_non_vxi(command, timeout)

		return stb

	# noinspection PyTypeChecker
	def _write_and_query_opc(self, cmd: str, timeout: int) -> StatusByte:
		"""Internal method to write a command followed by query_opc().
		Used for opc-synchronization if the mode is set to WaitForOpcMode.opc_query or the session is not-vxi.
		Timeout value 0 means the OPC timeout is used."""
		old_tout = self.visa_timeout

		# Change VISA Timeout if necessary
		if old_tout != timeout:
			self.visa_timeout = timeout
		try:
			# try-catch to set the VISA timeout back
			self.write(cmd)
			self.query_opc()
		finally:
			if old_tout != timeout:
				self.visa_timeout = old_tout
		return self._query_stb()

	def clear_status_after_query_with_opc(self) -> bool:
		"""Returns true, if the opc-sync queries require status clearing afterward."""
		if self.vxi_capable is False or self._opc_wait_mode is WaitForOpcMode.opc_query:
			return False
		if self.opc_sync_query_mechanism == OpcSyncQueryMechanism.standard:
			return True
		if self.opc_sync_query_mechanism == OpcSyncQueryMechanism.also_check_mav:
			return True
		return False

	def write_raw_bytes(self, data: bytes) -> None:
		"""Writes raw bytes to the instrument."""
		self._session.write_raw(data)

	def write(self, cmd: str) -> None:
		"""Writes command to the instrument."""
		if self.write_delay > 0:
			time.sleep(self.write_delay / 1000)
		add_tc = False

		if self._assure_write_with_tc and not cmd.endswith(self._term_char):
			add_tc = True

		if self.each_cmd_prefix:
			cmd = self.each_cmd_prefix + cmd

		if self.each_cmd_suffix:
			cmd = cmd + self.each_cmd_suffix

		cmd_bytes = cmd.encode(self.encoding)

		if add_tc:
			cmd_bytes += self._term_char.encode(self.encoding)
		self.write_raw_bytes(cmd_bytes)

	def _read_unknown_len(self, stream: StreamWriter, allow_chunk_events: bool, prepend_data: AnyStr = None) -> None:
		"""Reads data of unknown length to the provided WriteStream.
		The read is performed in an incremental chunk steps to optimize memory use (for NRP-Z session it is set to fixed self._data_chunk_size):
			- The first read is performed with the fixed size of 1024 bytes
			- The 2nd one reads 64 kBytes
			- The 3rd one reads 128 kBytes
			- The 4th one reads 256 kBytes and so on, with the max cap of self._data_chunk_size
		:param stream: [StreamWriter] target for the read data
		:param allow_chunk_events: [bool] if True, the method can send the chunk_events. If False, sending events is blocked.
		:param prepend_data: Optional[bytes or string] You can prepend this data to the beginning. It will be considered part of the first read chunk
		:return: read data [bytes or string], depending on the parameter binary."""
		with self._session.ignore_warning(pyvisa.constants.StatusCode.success_max_count_read):
			if prepend_data and isinstance(prepend_data, str):
				prepend_data = prepend_data.encode(self.encoding)
			chunk_ix = 0
			eot = False
			while not eot:
				if self.is_rsnrp_session():
					chunk_size = self._data_chunk_size
				else:
					if chunk_ix == 0:
						# First read, set 1024 bytes read size
						chunk_size = 1024
					elif chunk_ix == 1:
						chunk_size = 65536
					else:
						chunk_size *= 2
				if chunk_size > self._data_chunk_size:
					chunk_size = self._data_chunk_size
				chunk, self.last_status = self._session.visalib.read(self._session.session, chunk_size)
				if chunk_ix == 0 and prepend_data:
					chunk = prepend_data + chunk
				eot = not self._last_status_more_data_available()
				if not stream.binary:
					chunk = chunk.decode(self.encoding)
					if eot:
						chunk = chunk.rstrip(self._term_char)
						if self.strip_str_trailing_ws:
							chunk = chunk.rstrip()

				stream.write(chunk)
				if self.on_read_chunk_handler and allow_chunk_events:
					total_size = len(stream) if eot is True else None
					event_args = EventArgsChunk(stream.binary, chunk_ix, len(chunk), total_size, len(stream), eot, None, chunk if self.io_events_include_data else None)
					# noinspection PyCallingNonCallable
					self.on_read_chunk_handler(event_args)
				chunk_ix += 1

	def _last_status_more_data_available(self):
		"""Returns True, if the last status signaled that more data is available"""
		return self.last_status == pyvisa.constants.StatusCode.success_max_count_read

	def _read_str_no_events(self) -> str | None:
		"""Reads response from the instrument. The response is then trimmed for trailing LF. \n
		Sending of any read events is blocked."""
		if self.read_delay > 0:
			time.sleep(self.read_delay / 1000)
		stream = StreamWriter.as_string_var()
		self._read_unknown_len(stream, False)
		return stream.content

	def _query_str_no_events(self, query: str, allow_tout_error_narrow_down: bool = True) -> str | None:
		"""Queries the instrument and reads the response as string.
		The length of the string is not limited. The response is then trimmed for trailing LF.
		Sending of any read events is blocked. Use this method for all the service VisaSession queries."""
		response = ''
		self.write(query)
		try:
			response = self._read_str_no_events()
		except pyvisa.VisaIOError as e:
			context = f"Query '{query.rstrip(self._term_char)}'"
			if e.error_code == StatusCode.error_timeout:
				if allow_tout_error_narrow_down:
					self._narrow_down_io_tout_error(context + ' - ')
				else:
					raise TimeoutException(context)
			else:
				raise RsInstrException(context)
		return response

	def _query_str_no_events_timed(self, query: str, timeout: int) -> str:
		"""Queries the instrument and reads the response as string.
		The entered timeout sets the VISA timeout just for this call. You can suppress the timeout error.
		The length of the string is not limited. The response is then trimmed for trailing LF.
		Sending of any read events is blocked. Use this method for all the service VisaSession queries."""
		response = ''
		self.write(query)
		try:
			response = self._read_str_timed(timeout, False)
		except pyvisa.VisaIOError:
			self._narrow_down_io_tout_error(f"Query with timeout {timeout} ms '{query.rstrip(self._term_char)}' - ", timeout)
		return response

	def _read_str_timed(self, timeout: int, suppress_read_tout: bool = False) -> str | None:
		"""Reads response from the instrument with a VISA timeout temporarily set for the read.
		The VISA timeout is set back to the previous value before the method finishes even if an exception occurs.
		Sending of any read events is blocked."""
		old_visa_tout = self.visa_timeout
		if suppress_read_tout:
			try:
				if timeout != old_visa_tout:
					self.visa_timeout = timeout
				data = self._read_str_no_events()
				return data
			except pyvisa.VisaIOError as err:
				if err.error_code == pyvisa.constants.VI_ERROR_TMO:
					return None
				else:
					raise err
			finally:
				self.visa_timeout = old_visa_tout
		else:
			try:
				if timeout != old_visa_tout:
					self.visa_timeout = timeout
				data = self._read_str_no_events()
				return data
			finally:
				self.visa_timeout = old_visa_tout

	def read_str(self) -> str | None:
		"""Reads response from the instrument. The response is then trimmed for trailing LF."""
		if self.read_delay > 0:
			time.sleep(self.read_delay / 1000)
		stream = StreamWriter.as_string_var()
		self._read_unknown_len(stream, True)
		return stream.content

	def query_str(self, query: str) -> str:
		"""Queries the instrument and reads the response as string.
		The length of the string is not limited. The response is then trimmed for trailing LF."""
		response = ''
		self.write(query)
		try:
			response = self.read_str()
		except pyvisa.VisaIOError:
			self._narrow_down_io_tout_error(f"Query '{query.rstrip(self._term_char)}' - ")
		return response

	# noinspection PyTypeChecker
	def query_str_no_tout_err(self, query: str, tout: int) -> str:
		"""Same as query_str, but you can set the timeout just for this one call.
		If the timeout exception occurs, it is suppressed and the method returns Null"""
		response = None
		old_tout = self.visa_timeout
		try:
			self.visa_timeout = tout
			response = self.query_str(query)
		except (pyvisa.VisaIOError, StatusException):
			pass
		finally:
			self.visa_timeout = old_tout
		return response

	def write_with_opc(self, command: str, timeout: int = None) -> None:
		"""Sends command with OPC-sync.
		If you do not provide timeout, the method uses current opc_timeout."""
		self._write_and_wait_for_opc(command, False, timeout)

	# noinspection PyTypeChecker
	def query_str_with_opc(self, query: str, timeout: int = None, context: str = 'Query string with OPC') -> str:
		"""Query string with OPC synchronization.
		The response is trimmed for any trailing LF.
		If you do not provide timeout, the method uses current opc_timeout."""
		timeout = self._resolve_opc_timeout(timeout)
		if self.vxi_capable and self._opc_wait_mode is not WaitForOpcMode.opc_query:
			# For Vxi session, use the STB poll or SRQ wait and then read the response
			stb = self._write_and_wait_for_opc(query, True, timeout)
			self._check_msg_available_after_opc_wait(stb, query, timeout, context)
			response = self.read_str()
			return response
		else:
			# For non-Vxi sessions, use the longer VISA Timeout without the *OPC?
			# Same is valid for WaitForOpcMode.OpcQuery
			assert_query_has_qmark(query, 'Query with VISA timeout')
			self.write(query)
			old_tout = self.visa_timeout
			# Change VISA Timeout if necessary
			if old_tout != timeout:
				self.visa_timeout = timeout
			try:
				# try-catch to set the VISA timeout back
				response = self.read_str()
				if self._opc_wait_mode is WaitForOpcMode.opc_query:
					self.query_opc()
				return response
			finally:
				if old_tout != timeout:
					self.visa_timeout = old_tout

	def query_opc(self, timeout: int = 0) -> bool:
		"""Sends *OPC? query and reads the result.
		If you define timeout > 0, the VISA timeout is set to that value just for this method call."""
		if self.disable_opc_query:
			return True
		if timeout > 0:
			response = self._query_str_no_events_timed('*OPC?', timeout)
		else:
			response = self._query_str_no_events('*OPC?')
		return Conv.str_to_bool(response)

	def query_and_clear_esr(self) -> int:
		"""Sends *ESR? query and reads the result."""
		response = self._query_str_no_events('*ESR?')
		return int(response)

	def _check_msg_available_after_opc_wait(self, stb: StatusByte, query: str, timeout: int, context: str) -> None:
		"""Used internally after _StbPolling() to check if the message is available.
		Throws an exception in case of MAV not available."""
		if not self.vxi_capable:
			return
		if stb & StatusByte.message_available:
			return
		# Message not available
		context = context + f" Query '{query.rstrip(self._term_char)}'"
		if stb & StatusByte.error_queue_not_empty:
			# Instrument reports an error
			assert_no_instrument_status_errors(self.resource_name, self.query_all_syst_errors(), context)
		else:
			# Sometimes even if the StatusByte.MessageAvailable is false, the message is available.
			# Try to read the STB again
			stb = self._read_stb()
			if not stb & StatusByte.event_status_byte:
				# Instrument did not respond within the defined time
				throw_opc_tout_exception(self.opc_timeout, timeout, f'{context} No response from the instrument.')

	def error_in_error_queue(self) -> bool:
		"""Returns true, if error queue contains at least one error."""
		stb = self._query_stb()
		if stb & StatusByte.error_queue_not_empty:
			return True
		return False

	def reset_ese_sre(self) -> None:
		"""Resets the status of ESE and SRE registers to default values."""
		self._set_regs_ese_sre(self._opc_wait_mode)

	def write_bin_block(self, cmd: str, data_stream: StreamReader) -> None:
		"""Writes all the payload as binary data block to the instrument.
		The binary data header is added at the beginning of the transmission automatically.
		:param cmd: [str] SCPI command with which to send the data
		:param data_stream: [StreamReader] data provider for the payload"""
		data_size = len(data_stream)
		len_str = f'{data_size}'
		cmd = cmd.rstrip(self._term_char)
		if '#' in cmd:
			raise RsInstrException(
				f"Command '{cmd}' must be provided without the binary data header. "
				f"The method 'write_bin_block' composes and prepends the binary data header automatically.")
		if data_size <= self._std_bin_block_header_max_len:
			# Standard bin data header for sizes below 1E9 bytes, e.g.: '#512345'
			cmd_plus_header = f'{cmd}#{len(len_str)}{len_str}'.encode(self.encoding)
		else:
			# Big sizes bin data header: e.g.: '#(3000000000)'
			cmd_plus_header = f'{cmd}#({len_str})'.encode(self.encoding)

		if data_size <= self._data_chunk_size:
			# Write all in one step
			full_chunk = data_stream.read_as_binary(self.encoding)
			write_buf = cmd_plus_header + full_chunk
			if self._add_term_char_to_write_bin_block:
				write_buf += self._term_char_bin
			self.write_raw_bytes(write_buf)
			# Event sending
			if self.on_write_chunk_handler:
				event_args = EventArgsChunk(True, 0, data_size, data_size, data_size, True, 1, full_chunk if self.io_events_include_data else None)
				self.on_write_chunk_handler(event_args)
		else:
			# Write in chunks
			try:
				# Use finally to set the session send_end back to True
				self._session.send_end = False
				total_chunks = calculate_chunks_count(data_size, self._data_chunk_size)
				chunk_ix = 0
				if self.write_delay > 0:
					time.sleep(self.write_delay / 1000)
				# Write bin header
				self.write_raw_bytes(cmd_plus_header)
				# Write chunks
				while True:
					if len(data_stream) > self._data_chunk_size:
						#  Not the last segment
						chunk = data_stream.read_as_binary(self.encoding, self._data_chunk_size)
						self.write_raw_bytes(chunk)
						# Event sending
						if self.on_write_chunk_handler:
							event_args = EventArgsChunk(
								True, chunk_ix, self._data_chunk_size, data_size, data_size - len(data_stream), False, total_chunks, chunk if self.io_events_include_data else None)
							self.on_write_chunk_handler(event_args)
					else:
						# Last segment, indicate end of message again
						chunk = data_stream.read_as_binary(self.encoding)
						if self._add_term_char_to_write_bin_block:
							# Append LF
							self.write_raw_bytes(chunk)
							self._session.send_end = True
							self.write_raw_bytes(self._term_char_bin)
						else:
							self._session.send_end = True
							self.write_raw_bytes(chunk)

						# Event sending
						if self.on_write_chunk_handler:
							event_args = EventArgsChunk(True, chunk_ix, len(chunk), data_size, data_size, True, total_chunks, chunk if self.io_events_include_data else None)
							self.on_write_chunk_handler(event_args)
						break
					chunk_ix += 1
			finally:
				self._session.send_end = True

	def _parse_bin_data_header(self, exc_if_not_bin: bool) -> Tuple[ReadDataType, str, int]:
		"""Parses the binary data block and returns the expected length of the following data block. \n
		:param exc_if_not_bin: [bool] if True, the method throws exception in case the data is not binary.
		:return: read_data_type: [ReadDataType], parsed_header: [string], bin_data_len: [integer]"""
		length = -1
		if self.read_delay > 0:
			time.sleep(self.read_delay / 1000)

		char = self._session.read_bytes(1, break_on_termchar=True)
		if char == b'#':
			# binary transfer
			char = self._session.read_bytes(1, break_on_termchar=True)
			if char == b'0':
				data_type = ReadDataType.bin_unknown_len
				return data_type, '#0', -1
			if char == b'(':
				# format for big lengths i.e. > 1E9 bytes: '#(1234567890123)...'
				data_type = ReadDataType.bin_known_len
				len_str = (self.read_up_to_char(b')', 100)[:-1]).decode(self.encoding)
				whole_hdr = '#(' + len_str + ')'
				length = int(len_str)
				return data_type, whole_hdr, length

			# classic format for < 1E9 bytes: '#9123456789...'
			data_type = ReadDataType.bin_known_len
			len_of_len = int(char)
			len_str = self._session.read_bytes(len_of_len).decode(self.encoding)
			length = int(len_str)
			whole_hdr = '#' + char.decode(self.encoding) + len_str
			return data_type, whole_hdr, length

		data_type = ReadDataType.ascii
		if char == self._term_char_bin:
			data_type = ReadDataType.null
		if self.vxi_capable:
			# For Vxi session, to be sure, check whether there are more chars in the read buffer
			stb = self._read_stb()
			if stb & StatusByte.message_available:
				data_type = ReadDataType.ascii
		whole_hdr = char.decode(self.encoding)
		if exc_if_not_bin:
			if data_type == ReadDataType.null:
				throw_bin_block_unexp_resp_exception(self.resource_name, self._term_char)
			# Read 20 more characters to compose a better exception message
			whole_hdr += self.read_up_to_char(self._term_char_bin, 20).decode(self.encoding)
			if self.last_status == pyvisa.constants.StatusCode.success_max_count_read:
				self._flush_junk_data()
			throw_bin_block_unexp_resp_exception(self.resource_name, whole_hdr)
		return data_type, whole_hdr, length

	def get_bin_data_length(self, query: str) -> int | None:
		"""Returns only the length binary data header, and discards the actual data.
		Any timeout error is suppressed, and the method returns None instead.
		Warning!!! - for non-VXI sessions (SOCKET, ASRL) this method transfers the entire file to the control PC, which might take a long time."""
		if self.vxi_capable and self._opc_wait_mode != WaitForOpcMode.opc_query:
			# For Vxi session, use the STB poll and read the header
			stb = self._write_and_wait_for_opc(query, True, 0)
			try:
				self._check_msg_available_after_opc_wait(stb, query, 0, 'get_bin_data_length')
			except StatusException:
				return None
			data_type, header, length = self._parse_bin_data_header(True)
			self.clear()
			self.clear_before_read()
			return length
		else:
			with StreamWriter.as_forget() as stream:
				old_timeout = self.visa_timeout
				try:
					self.visa_timeout = 2000
					self.query_bin_block(query, stream, True)
				except StatusException:
					return None
				finally:
					self.visa_timeout = old_timeout
				length = stream.written_len
		return length

	def read_bin_block(self, stream: StreamWriter, exc_if_not_bin: bool) -> None:
		"""Reads binary data block to the provided stream. \n
		:param stream: [StreamWriter] target for the read data. Can be string, bytes, or a file
		:param exc_if_not_bin: if True, the method throws exception if the received data is not binary"""
		data_type, header, length = self._parse_bin_data_header(exc_if_not_bin)
		if data_type == ReadDataType.ascii:
			stream.switch_to_string_data(self.encoding)
			self._read_unknown_len(stream, True, header)
		elif data_type == ReadDataType.null:
			# No data, consider it ASCII. Change the stream type to ASCII and return empty string
			stream.switch_to_string_data(self.encoding)
		elif data_type == ReadDataType.bin_unknown_len:
			if not self.vxi_capable:
				raise RsInstrException(f'Non-Vxi11 sessions can not read binary data blocks of unknown length.')
			self._read_unknown_len(stream, True)
		elif length == 0:
			self._flush_junk_data()
		else:
			self._read_bin_block_known_len(stream, length)

	def read_all_bytes(self, stream: StreamWriter) -> None:
		"""Reads all the data from the instrument as bytes to the provided stream."""
		if not self.vxi_capable:
			raise RsInstrException(f'Non-Vxi11 sessions can not read binary data blocks of unknown length.')
		self._read_unknown_len(stream, True)

	def _read_bin_block_known_len(self, stream: StreamWriter, length: int) -> None:
		"""Reads binary data of defined length. All remaining data above the length are disposed of. \n
		:param stream: [StreamWriter] target for the read data. Can be string, bytes, or a file
		:param length: [int] expected length of the data"""
		# Use try-catch to switch the termination character back ON in case of an exception (for non-Vxi sessions)
		try:
			# Binary transmission, for non-Vxi session, set the termination character to OFF
			if not self.vxi_capable:
				self._session.read_termination = False
			# Binary data of known length
			left_to_read = length
			self.last_status = pyvisa.constants.StatusCode.success
			with self._session.ignore_warning(pyvisa.constants.StatusCode.success_max_count_read):
				chunk_ix = 0
				total_chunks = calculate_chunks_count(length, self._data_chunk_size)
				while len(stream) < length:
					chunk_size = min(self._data_chunk_size, left_to_read)
					chunk, self.last_status = self._session.visalib.read(self._session.session, chunk_size)
					left_to_read -= len(chunk)
					stream.write(chunk)
					if self.on_read_chunk_handler:
						event_args = EventArgsChunk(True, chunk_ix, chunk_size, length, len(stream), left_to_read == 0, total_chunks, chunk if self.io_events_include_data else None)
						self.on_read_chunk_handler(event_args)
					chunk_ix += 1
			if self._last_status_more_data_available():
				if not self.vxi_capable:
					self._session.read_termination = self._term_char
				self._flush_junk_data()
		finally:
			# Make sure that in any case the self._session.read_termination is ON again for non-Vxi sessions
			if not self.vxi_capable:
				self._session.read_termination = self._term_char

	def query_bin_block(self, query: str, stream: StreamWriter, exc_if_not_bin: bool = True) -> None:
		"""Query binary data block and returns it as byte data. \n
		:param query: [str] query to send to the instrument
		:param stream: [StreamWriter] target for the read data. Can be string, bytes, or a file
		:param exc_if_not_bin: [Boolean] if True, the method throws exception if the received data is not binary"""
		self.write(query)
		try:
			self.read_bin_block(stream, exc_if_not_bin)
		except pyvisa.VisaIOError:
			self._narrow_down_io_tout_error(f"Query bin block '{query.rstrip(self._term_char)}' - ")
		return

	def query_bin_block_with_opc(self, query: str, stream: StreamWriter, exc_if_not_bin: bool = True, timeout: int = None) -> None:
		"""Query binary data block with OPC and returns it as byte data.
		:param query: [str] query to send to the instrument
		:param stream: [StreamWriter] target for the read data. Can be string, bytes, or a file
		:param exc_if_not_bin: [Boolean] if True, the method throws exception if the received data is not binary
		:param timeout: Optional[Integer] timeout for the operation. If you skip it, the method uses the current opc timeout."""
		timeout = self._resolve_opc_timeout(timeout)
		if self.vxi_capable and self._opc_wait_mode != WaitForOpcMode.opc_query:
			# For Vxi session, use the STB poll and read the response
			stb = self._write_and_wait_for_opc(query, True, timeout)
			self._check_msg_available_after_opc_wait(stb, query, timeout, 'query_bin_block_with_opc')
			self.read_bin_block(stream, exc_if_not_bin)
		else:
			# For non-Vxi session, use the longer VISA Timeout without the *OPC
			assert_query_has_qmark(query, 'query_bin_block_with_opc')
			self.write(query)
			old_visa_timeout = self.visa_timeout
			# Change VISA Timeout if necessary
			if old_visa_timeout != timeout:
				self.visa_timeout = timeout
			try:
				# try-catch to set the VISA timeout back
				self.read_bin_block(stream, exc_if_not_bin)
				if self._opc_wait_mode == WaitForOpcMode.opc_query:
					self.query_opc()
			finally:
				# Change VISA Timeout back if necessary
				if old_visa_timeout != timeout:
					self.visa_timeout = old_visa_timeout

	def read_up_to_char(self, stop_chars: bytes, max_cnt: int) -> bytes:
		"""Reads until one of the stop_chars is read or the max_cnt is reached, or EOT is detected.
		Returns the read data including the stop character."""
		response = b''
		for i in range(max_cnt):
			char, self.last_status = self._session.visalib.read(self._session.session, 1)
			response += char
			if char in stop_chars:
				break
			if self.last_status != pyvisa.constants.StatusCode.success_max_count_read:
				break
		return response

	def go_to_local(self, mixed_mode: bool) -> None:
		"""Puts the instrument into local state.
		By default, the method uses a mechanism to keep the instrument in a mixed mode: remote and local.
		That means, you can remote-control your instrument, and at the same time it still allows manual control.
		Set the mixed_mode to False, if you want your instrument to go to remote mode as soon as it receives the first remote command."""
		if self.vxi_capable:
			if mixed_mode:
				self._session.control_ren(pyvisa.constants.RENLineOperation.deassert_gtl)
			else:
				self._session.control_ren(pyvisa.constants.RENLineOperation.address_gtl)
		else:
			self.write("&GTL")

	def go_to_remote(self) -> None:
		"""Puts the instrument into remote state."""
		if self.vxi_capable:
			self._session.control_ren(pyvisa.constants.RENLineOperation.asrt_address)
		else:
			self.write("&GTR")

	def get_session_handle(self) -> object:
		"""Returns the underlying pyvisa session."""
		return self._session

	def close(self) -> None:
		"""Closes the Visa session.
		If the object was created with the direct session input, the session is not closed."""
		if not self.reusing_session:
			self._session.close()

	# Events


class EventArgsChunk:
	"""Event arguments for chunk io event."""

	def __init__(
			self,
			binary: bool,
			chunk_ix: int,
			chunk_size: int,
			total_size: int,
			transferred_size: int,
			end_of_transfer: bool,
			total_chunks: int | None,
			data: AnyStr = None):

		self.binary = binary
		self.chunk_ix = chunk_ix
		self.total_chunks = total_chunks
		self.chunk_size = chunk_size
		self.transferred_size = transferred_size
		self.total_size = total_size
		self.end_of_transfer = end_of_transfer
		self.data = data

	def __str__(self):
		if self.binary:
			type_info = 'binary'
		else:
			type_info = 'ascii'
		if not self.total_chunks:
			chunk_info = f' chunk nr. {self.chunk_ix + 1}'
		elif self.total_chunks > 1:
			chunk_info = f' chunk nr. {self.chunk_ix + 1}/{self.total_chunks}'
		else:
			chunk_info = ' chunk nr. 1/1'
		eot = ' (EOT)' if self.end_of_transfer else ''
		result = \
			f'EventArgsChunk {type_info},{chunk_info}, {size_to_kb_mb_string(self.chunk_size, True)}, ' \
			f'sum {size_to_kb_mb_string(self.transferred_size, True)} / {size_to_kb_mb_string(self.total_size, True) if self.total_size else "<N.A.>"}{eot}.'
		return result



from typing import Any
from ..Internal.Core import Core


class IviDirectIo:
    """Direct IO class required by the IVI-Python standard."""

    def __init__(self, core: Core):
        self._core = core

    @property
    def session(self) -> Any:
        """Returns the underlying instrument session, usually a VISA session."""
        return self._core.io.get_session_handle()

    @property
    def io_timeout_ms(self) -> int:
        """Return the timeout of the underlying instrument session in milliseconds."""
        return self._core.io.visa_timeout

    @io_timeout_ms.setter
    def io_timeout_ms(self, timeout_ms: int) -> None:
        """Sets the timeout of the underlying instrument session in milliseconds."""
        self._core.io.visa_timeout = timeout_ms

    def read_bytes(self) -> bytes:
        """Read a complete response from the instrument as bytes.
        The response message terminator is not included in the bytes object."""
        return self._core.io.read_all_bytes()
 
    def read_string(self) -> str:
        """Read a complete response from the instrument as a string.
        The response message terminator is not included in the string."""
        return self._core.io.read_str(block_check_status = True)
 
    def write_bytes(self, data: bytes) -> None:
        """Write bytes to the instrument followed by the normal message termination sequence.
        For IEEE 488.2 instruments the termination sequence is typically a line feed character with END asserted.
        The bytes must include a complete instrument message. The driver adds the termination sequence."""
        self._core.io.write_raw_bytes(data)
 
    def write_string(self, data: str) -> None:
        """Write a string to the instrument followed by the normal message termination sequence.
        For IEEE 488.2 instruments the termination sequence is typically a line feed character with END asserted.
        The string must include a complete instrument message. The driver adds the termination sequence."""
        self._core.io.write(data, block_check_status = True)






"""Links a variable from a structure to a callback in the driver."""

from time import time
from typing import Dict, Callable

from .ArgSingle import ArgSingle
from .ArgSingleSuppressed import ArgSingleSuppressed
from .ArgLinkedEventArgs import ArgLinkedEventArgs
from .Utilities import get_plural_string
from .InstrumentErrors import RsInstrException


class InternalLinker(object):
	"""Class for:
		- cutting out suppressed arguments from a device response.
		- invoking a handler if the argument has InternalLinking defined.
		- holds dictionary of handlers where the dict_key is the InternalLinking string.
	Handlers registration/deleting is done with:
		- set_handler()
		- del_handler()
		- del_all_handlers()"""

	def __init__(self):
		self._handlers = {}

	def __str__(self):
		if len(self._handlers) == 0:
			return 'Linker, no handlers'
		return f"Linker, {get_plural_string('handler', len(self._handlers))}: {','.join(self._handlers)}"

	def set_handler(self, link_name: str, handler: Callable) -> Callable:
		"""Adds / Updates handler for the link_name.
		Returns the previous registered handler, or None if no handler was registered before."""
		previous = None if link_name not in self._handlers else self._handlers[link_name]
		self._handlers[link_name] = handler
		return previous

	def del_handler(self, link_name: str) -> Callable:
		"""Deletes handler for the link_name.
		Returns the deleted handler, or None if none existed."""
		current = None if link_name not in self._handlers else self._handlers[link_name]
		if current:
			del self._handlers[link_name]
		return current

	def del_all_handlers(self) -> int:
		"""Deletes all the handlers.
		Returns number of deleted links."""
		count = len(self._handlers)
		self._handlers = {}
		return count

	def cut_from_response_string(self, arg: ArgSingleSuppressed, response: str, context: str) -> str:
		"""Takes the string 'response', removes the suppressed argument value from it and returns the rest.
		The cut-out part is sent via handler if the internal linking exists for that argument exists."""
		result = ''
		if arg.argument_ix is None:
			raise RsInstrException(f'Argument has argument_ix attribute not assigned (equals None). Argument: {arg}')
		if arg.argument_ix != 0:
			raise RsInstrException(f'Only arguments with index 0 can be suppressed. Argument: {arg}')
		if arg.is_open_list:
			raise RsInstrException(f'Open List arguments can not be suppressed. Argument: {arg}')
		repetition = 0
		i = 0
		for c in response:
			if c == ',':
				repetition += 1
			if repetition == arg.repetition:
				break
			i += 1

		suppressed_part = response[0:i]
		i += 1
		if i < len(response):
			result = response[i:]

		self.invoke_single_intern_link(arg, context, suppressed_part)
		return result

	def invoke_single_intern_link(self, arg: ArgSingleSuppressed | ArgSingle, context: str, value: str) -> None:
		"""Invokes the registered handler for the internal linked argument."""
		if arg.intern_link and arg.intern_link in self._handlers:
			event_args = ArgLinkedEventArgs(arg.intern_link, arg.name, value, context, time())
			self._handlers[arg.intern_link](event_args)

	def invoke_struct_intern_links(self, struct, args: Dict, context: str) -> None:
		"""Invokes handler for each of the Structure arguments that have internal linking."""
		for arg in args.values():
			if arg.intern_link and arg.intern_link in self._handlers:
				event_args = ArgLinkedEventArgs(arg.intern_link, arg.name, getattr(struct, arg.name), context, time())
				self._handlers[arg.intern_link](event_args)



"""Singleton class that holds values reachable from anywhere in the Internal package.
The first instantiation creates the object, the consequent ones return the same object."""


class Properties(object):
    """Class holding global properties."""
    scpi_quotes = "'"

    @classmethod
    def reset(cls):
        """Sets all the variables to default."""
        cls.scpi_quotes = "'"



"""See the class docstring."""

from . import Utilities
from .ArgStruct import ArgStruct
from .InstrumentErrors import RsInstrException


class ArgStructStringParser:
	"""Class for parsing a response from the instrument to an output structure of arguments.
	It is used by the ArgStructList class for filling structures with return values."""

	def __init__(self, struct, value: str):
		self.struct = struct
		self.elements = value.split(',')
		self.count = len(self.elements)
		self.position = 0

	@property
	def remaining(self) -> int:
		"""Remaining items to parse."""
		return self.count - self.position

	def to_scalar_value(self, arg: ArgStruct):
		"""Parses the current element to a scalar argument."""
		assert arg.data_type.is_scalar, f'to_scalar_value() method only works with scalar values. Data type: {arg.data_type}'
		if self.position >= self.count:
			raise RsInstrException(
				f"Cannot parse a scalar value to structure argument. Response contains only {self.count} elements, "
				f"argument '{arg.name}' has position {self.position + 1}.\n"
				f"Response (commas replaced by new lines):\n" + Utilities.truncate_string_from_end('\n'.join(self.elements), 1000))
		string = self.elements[self.position]
		value = arg.conv_from_scpi_string.get_one_element_value(string)
		setattr(self.struct, arg.name, value)
		self.position += 1

	def to_list_value(self, arg: ArgStruct, increase_pos: bool, offset: int, count: int, period: int, cycles: int) -> None:
		"""Parses more elements to the list argument - slicing."""
		assert arg.data_type.is_list, f'to_list_value method only works with list values. Data type: {arg.data_type}'
		if cycles < 0:
			cycles = self.remaining // period
		if self.position >= self.count:
			raise RsInstrException(
				f"Cannot parse an list value to the argument '{arg.name}', "
				f"because the element position {self.position} is over the parsed list length {self.count}")
		if (self.position + offset + count) > self.count:
			raise RsInstrException(
				f"Cannot parse the whole list value to the argument '{arg.name}', because the element position {self.position} "
				f"plus the argument offset {offset} and argument length {count} would be over the parsed list length {self.count}")

		result = []
		for cycle in range(cycles):
			start_ix = self.position + (cycle * period) + offset
			for i in range(count):
				result.append(arg.conv_from_scpi_string.get_one_element_value(self.elements[start_ix + i]))

		if increase_pos:
			self.position += count

		setattr(self.struct, arg.name, result)



"""Root class for remote-controlling instrument with SCPI commands."""

import threading
from typing import List, Tuple, ClassVar
from datetime import datetime, timedelta

from .Fixed_Files.Events import Events
from .Fixed_Files.ivi_direct_io import IviDirectIo
from .Fixed_Files.ivi_utility import IviUtility
from .Internal.Core import Core
from .Internal.Conversions import BinFloatFormat, BinIntFormat
from RsInstrument.Internal.InstrumentSettings import OpcSyncQueryMechanism
from .Internal import Conversions as Conv
from .Internal.VisaSession import VisaSession
from .Internal.ScpiLogger import ScpiLogger
from .Internal.Utilities import trim_str_response
from .Internal.GlobalData import GlobalData
from .Internal.ContextManagers import InstrErrorSuppressor, VisaTimeoutSuppressor


class RsInstrument:
	"""Root class for remote-controlling instrument with SCPI commands."""
	_driver_version_const = '1.124.0.122'
	_driver_options_const = "SupportedInstrModels = All Rohde & Schwarz Instruments, SupportedIdnPatterns = Rohde\\s*(-|&)\\s*Schwarz/Hameg, SimulationIdnString = Rohde&Schwarz*SimulationDevice*100001*" + _driver_version_const
	# noinspection PyClassVar
	_global_logging_relative_timestamp: ClassVar[datetime] = None
	_global_logging_target_stream: ClassVar = None
	_global_logging_relative_timestamp_of_first_entry: ClassVar = False

	def __init__(
			self, resource_name: str, id_query: bool = True, reset: bool = False, options: str = None, direct_session: object = None):
		"""Initializes new RsInstrument session. \n

		:param resource_name: VISA resource name, e.g. 'TCPIP::192.168.2.1::INSTR'
		:param id_query: if True, the instrument's model name is verified against the models supported by the driver and eventually throws an exception
		:param reset: Resets the instrument (sends *RST) command and clears its status syb-system
		:param direct_session: Another driver object or pyVisa object to reuse the session instead of opening a new session
		:param options: string tokens alternating the driver settings. More tokens are separated by comma.

		Parameter options tokens examples:
			- ``Simulate=True`` - starts the session in simulation mode. Default: ``False``
			- ``SelectVisa=socketio`` - uses no VISA implementation for socket connections - you do not need any VISA-C installation
			- ``SelectVisa=rs`` - prefers usage of RohdeSchwarz VISA
			- ``SelectVisa=ni`` - prefers usage of National Instruments VISA
			- ``SelectVisa=pyvisa-py`` - prefers usage of python VISA backend pyvisa-py
			- ``Profile = HM8123`` - setting profile fitting the specific non-standard instruments. Available values: HM8123, CMQ, ATS, Minimal. Default: ``none``
			- ``OpenTimeout=5000`` - sets timeout used at the session opening. This timeout is only used in waiting for a locked session to be freed. Default: ``2000ms``
			- ``ExclusiveLock=True`` - opens the session with exclusive lock on the VISA level. Default: ``False``
			- ``QueryInstrumentStatus = False`` - same as ``driver.utilities.instrument_status_checking = False``. Default: ``True``
			- ``WriteDelay = 20, ReadDelay = 5`` - introduces delay of 20ms before each write and 5ms before each read. Default: ``0ms`` for both
			- ``TerminationCharacter = "\\r"`` - sets the termination character for reading. Default: ``\\n`` (LineFeed or LF)
			- ``AssureWriteWithTermChar = True`` - makes sure each command/query is terminated with termination character. Default: Interface dependent
			- ``AddTermCharToWriteBinBlock = True`` - adds one additional LF to the end of the binary data (some instruments require that). Default: ``False``
			- ``DataChunkSize = 10E3`` - maximum size of one write/read segment. If transferred data is bigger, it is split to more segments. Default: ``1E7`` bytes
			- ``OpcTimeout = 10000`` - same as driver.utilities.opc_timeout = 10000. Default: ``30000ms``
			- ``VisaTimeout = 5000`` - same as driver.utilities.visa_timeout = 5000. Default: ``10000ms``
			- ``ViClearExeMode = Disabled`` - viClear() execution mode. Default: ``execute_on_all``
			- ``OpcQueryAfterWrite = True`` - same as driver.utilities.opc_query_after_write = True. Default: ``False``
			- ``OpcWaitMode = OpcQuery`` - mode for all the opc-synchronized write/reads. Other modes: StbPolling, StbPollingSlow, StbPollingSuperSlow. Default: ``StbPolling``
			- ``StbInErrorCheck = False`` - if true, the driver checks errors with *STB? If false, it uses SYST:ERR?. Default: ``True``
			- ``SkipStatusSystemSettings = False`` - some instruments do not support full status system commands. In such case, set this value to True. Default: ``False``
			- ``SkipClearStatus = True`` - set to True for instruments that do not support *CLS command. Default: ``False``
			- ``DisableOpcQuery = True`` - set to True for instruments that do not support *OPC? query. Default: ``False``
			- ``EachCmdAsQuery = True``, set to True, for instruments that always return answer. Default: ``false``
			- ``CmdIdn = ID?`` - defines which SCPI command to use for identification query. Use '<none>' string to skip identification query at the init. Default: ``*IDN?``
			- ``CmdReset = RT`` - defines which SCPI command to use for reset. Default: ``*RST``
			- ``VxiCapable = false`` - you can force a session to a VXI-incapable. Default: <interface-dependent>
			- ``Encoding = utf-8`` - setting of encoding for strings into bytes and vice versa. Default: ``charmap``
			- ``OpcSyncQueryMechanism = AlsoCheckMav`` - setting of mechanism for OPC-synchronized queries. Default: ``OnlyCheckMavErrQueue``
			- ``FirstCmds = *CLS`` - first command(s) to sent after init. Separated more commands/queries with ';;'. Default: ````
			- ``EachCmdPrefix = lf`` - this prefix is added to the beginning of each command sent to the instrument. Default: ````
			- ``EachCmdSuffix = cr`` - this suffix is added to the end of each command sent to the instrument. Default: ````
			- ``StripStringTrailingWhitespaces = True`` - use it to strip white spaces from string query responses. Default: ``False``
			- ``LoggingMode = On`` - sets the logging status right from the start. Possible values: On | Off | Error. Default: ``Off``
			- ``LoggingName = 'MyDevice'`` - sets the name to represent the session in the log entries. Default: ``<resource_name>``
			- ``LoggingFormat = 'PAD_LEFT12(%START_TIME%) PAD_LEFT25(%DEVICE_NAME%) PAD_LEFT12(%DURATION%) %SCPI_COMMAND%'`` - sets the format of the log entries. Default: ``PAD_LEFT12(%START_TIME%) PAD_LEFT25(%DEVICE_NAME%) PAD_LEFT12(%DURATION%)  %LOG_STRING_INFO%: %LOG_STRING%``
			- ``LogToGlobalTarget = True`` - sets the logging target to the class-property previously set with RsInstrument.set_global_logging_target() Default: ``False``
			- ``LoggingToConsole = True`` - immediately starts logging to the console. Default: False
			- ``LoggingToUdp = True`` - immediately starts logging to the UDP port. Default: False
			- ``LoggingUdpPort = 49200`` - UDP port to log to. Default: 49200
			- ``LoggingRelativeTimeOfFirstEntry = True`` - Logging starts with relative time set to the first log entry, which causes the first start time to be '00:00:00.000'. Default: False
"""

		GlobalData.bounded_class = RsInstrument
		self._core = Core(
			resource_name,
			id_query,
			reset,
			driver_options=RsInstrument._driver_options_const,
			user_options=options,
			direct_session=direct_session,
			called_from_driver=False)
		self._core.driver_version = RsInstrument._driver_version_const

		# Custom interfaces
		self._events = Events(self._core)
		self._ivi_direct_io = IviDirectIo(self._core)
		self._ivi_utility = IviUtility(self._core)

	@classmethod
	def from_existing_session(cls, session: object, options: str = None) -> 'RsInstrument':
		"""Creates a new RsInstrument object with the entered 'session' reused.
		:param session: can be another driver or a direct pyvisa session.
		:param options: string tokens alternating the driver settings. More tokens are separated by comma."""
		GlobalData.bounded_class = RsInstrument
		resource_name = None
		if hasattr(session, 'resource_name'):
			resource_name = getattr(session, 'resource_name')
		# noinspection PyTypeChecker
		return cls(resource_name, False, False, options, session)

	@classmethod
	def set_global_logging_target(cls, target) -> None:
		"""Sets global common target stream that each instance can use. To use it, call the following: io.logger.set_logging_target_global().
		If an instance uses global logging target, it automatically uses the global relative timestamp (if set).
		You can set the target to None in order to invalidate it."""
		cls._global_logging_target_stream = target

	@classmethod
	def get_global_logging_target(cls):
		"""Returns global common target stream."""
		return cls._global_logging_target_stream

	@classmethod
	def set_global_logging_relative_timestamp(cls, timestamp: datetime) -> None:
		"""Sets global common relative timestamp for log entries. To use it, call the following: io.logger.set_relative_timestamp_global()"""
		cls._global_logging_relative_timestamp = timestamp

	@classmethod
	def set_global_logging_relative_timestamp_now(cls) -> None:
		"""Sets global common relative timestamp for log entries to this moment.
		To use it, call the following: io.logger.set_relative_timestamp_global()."""
		cls._global_logging_relative_timestamp = datetime.now()

	@classmethod
	def set_global_logging_relative_time_of_first_entry(cls) -> None:
		"""This method sets the global flag, that takes the very first log entry start time of any global instance as a global relative timestamp.
		This means, after this call, the first instance that logs an entry set the relative timestamp for all global target instances,
		and begins with the start timestamp 0:00:00.000"""
		cls._global_logging_relative_timestamp_of_first_entry = True

	@classmethod
	def clear_global_logging_relative_timestamp(cls) -> None:
		"""Clears the global relative timestamp. After this, all the instances using the global relative timestamp continue logging with the absolute timestamps."""
		# noinspection PyTypeChecker
		cls._global_logging_relative_timestamp = None

	@classmethod
	def get_global_logging_relative_timestamp(cls) -> datetime | None:
		"""Returns global common relative timestamp for log entries."""
		return cls._global_logging_relative_timestamp

	@classmethod
	def get_driver_version(cls) -> str:
		"""Returns the RsInstrument package version"""
		return cls._driver_version_const

	def __str__(self):
		"""String representation of the object."""
		if self._core.io:
			return f"RsInstrument session '{self._core.io.resource_name}'"
		else:
			return f"RsInstrument with session closed"

	def __enter__(self) -> 'RsInstrument':
		"""Stuff to do when entering the context.
		Currently, this method does nothing, and returns the RsInstrument object."""
		return self

	def __exit__(self, exc_type, value, traceback):
		"""Stuff to do when leaving the context.
		Closes the opened session."""
		self.close()

	def close(self) -> None:
		"""Closes the active RsInstrument session"""
		self._core.io.close()

	@staticmethod
	def list_resources(expression: str = '?*::INSTR', visa_select: str = None) -> List[str]:
		"""Finds all the resources defined by the expression.

		* '?*' - matches all the available instruments
		* 'USB::?*' - matches all the USB instruments
		* 'TCPIP::192?*' - matches all the LAN instruments with the IP address starting with 192

		:param expression: see the examples in the function
		:param visa_select: optional parameter selecting a specific VISA. Examples: '@ivi', '@rs'
		"""
		rm = VisaSession.get_resource_manager(visa_select)
		resources = rm.list_resources(expression)
		rm.close()
		# noinspection PyTypeChecker
		return resources

	def get_session_handle(self):
		"""Returns the underlying pyvisa session"""
		return self._core.get_session_handle()

	@property
	def events(self) -> Events:
		"""Interface for event handlers, see :ref:`here <Events>` """
		return self._events

	@property
	def ivi_direct_io(self) -> IviDirectIo:
		"""Interface for IVI standard Direct IO, see :ref:`here <IviDirectIo>` """
		return self._ivi_direct_io

	@property
	def ivi_utility(self) -> IviUtility:
		"""Interface for IVI standard Utility, see :ref:`here <IviUtility>` """
		return self._ivi_utility

	@property
	def logger(self) -> ScpiLogger:
		"""Scpi Logger interface, see :ref:`here <Logger>` """
		return self._core.io.logger

	# Utilities part follows - copy it manually from the Utilities.py file
	@property
	def driver_version(self) -> str:
		"""Returns the RsInstrument package version"""
		return self._core.driver_version

	@property
	def idn_string(self) -> str:
		"""Returns instrument's identification string - the response on the SCPI command *IDN?"""
		return self._core.io.idn_string

	@idn_string.setter
	def idn_string(self, idn_string: str) -> None:
		"""Sets instrument's identification string - the response on the SCPI command *IDN?"""
		self._core.io.idn_string = idn_string

	@property
	def manufacturer(self) -> str:
		"""Returns manufacturer of the instrument"""
		return self._core.io.manufacturer

	@property
	def full_instrument_model_name(self) -> str:
		"""Returns the current instrument's full name e.g. 'FSW26'"""
		return self._core.io.full_model_name

	@property
	def instrument_model_name(self) -> str:
		"""Returns the current instrument's family name e.g. 'FSW'"""
		return self._core.io.model

	@property
	def supported_models(self) -> List[str]:
		"""Returns a list of the instrument models supported by this instrument driver"""
		return self._core.supported_instr_models

	@property
	def instrument_firmware_version(self) -> str:
		"""Returns instrument's firmware version"""
		return self._core.io.firmware_version

	@property
	def instrument_serial_number(self) -> str:
		"""Returns instrument's serial_number"""
		return self._core.io.serial_number

	def query_opc(self, timeout: int = 0) -> int:
		"""SCPI command: *OPC?
		Queries the instrument's OPC bit and hence it waits until the instrument reports operation complete.
		If you define timeout > 0, the VISA timeout is set to that value just for this method call."""
		return self._core.io.query_opc(timeout)

	@property
	def instrument_status_checking(self) -> bool:
		"""Sets / returns Instrument Status Checking.
		When True (default is True), all the driver methods and properties are sending "SYSTem:ERRor?"
		at the end to immediately react on error that might have occurred.
		We recommend keeping the state checking ON all the time. Switch it OFF only in rare cases when you require maximum speed.
		The default state after initializing the session is ON."""
		return self._core.io.query_instr_status

	@instrument_status_checking.setter
	def instrument_status_checking(self, value) -> None:
		"""Sets / returns Instrument Status Checking.
		When True (default is True), all the driver methods and properties are sending "SYSTem:ERRor?"
		at the end to immediately react on error that might have occurred.
		We recommend keeping the state checking ON all the time. Switch it OFF only in rare cases when you require maximum speed.
		The default state after initializing the session is ON."""
		self._core.io.query_instr_status = value

	def check_status(self) -> None:
		"""Throws InstrumentStatusException in case of an error in the instrument's error queue.
		The status checking is performed always, independent of the property 'instrument_status_checking'.
		Also, the property ScpiLogger.log_status_check_ok is ignored, and the Status check is always logged."""
		self._core.io.check_status_always(True)

	@property
	def encoding(self) -> str:
		"""Returns string<=>bytes encoding of the session."""
		return self._core.io.encoding

	@encoding.setter
	def encoding(self, value: str) -> None:
		"""Sets string<=>bytes encoding of the session."""
		self._core.io.encoding = value

	@property
	def opc_query_after_write(self) -> bool:
		"""Sets / returns Instrument *OPC? query sending after each command write.
		When True, (default is False) the driver sends *OPC? every time a write command is performed.
		Use this if you want to make sure your sequence is performed command-after-command."""
		return self._core.io.opc_query_after_write

	@opc_query_after_write.setter
	def opc_query_after_write(self, value) -> None:
		"""Sets / returns Instrument *OPC? query sending after each command write.
		When True, (default is False) the driver sends *OPC? every time a write command is performed.
		Use this if you want to make sure your sequence is performed command-after-command."""
		self._core.io.opc_query_after_write = value

	@property
	def bin_float_numbers_format(self) -> BinFloatFormat:
		"""Sets / returns format of float numbers when transferred as binary data"""
		return self._core.io.bin_float_numbers_format

	@bin_float_numbers_format.setter
	def bin_float_numbers_format(self, value: BinFloatFormat) -> None:
		"""Sets / returns format of float numbers when transferred as binary data"""
		self._core.io.bin_float_numbers_format = value

	@property
	def bin_int_numbers_format(self) -> BinIntFormat:
		"""Sets / returns format of integer numbers when transferred as binary data"""
		return self._core.io.bin_int_numbers_format

	@bin_int_numbers_format.setter
	def bin_int_numbers_format(self, value: BinIntFormat) -> None:
		"""Sets / returns format of integer numbers when transferred as binary data"""
		self._core.io.bin_int_numbers_format = value

	def clear_status(self) -> None:
		"""Clears instrument's status system, the session's I/O buffers and the instrument's error queue"""
		return self._core.io.clear_status()

	def query_all_errors(self) -> List[str] | None:
		"""Queries and clears all the errors from the instrument's error queue.
		The method returns list of strings as error messages. If no error is detected, the return value is None.
		The process is: querying 'SYSTem:ERRor?' in a loop until the error queue is empty.
		If you want to include the error codes, call the query_all_errors_with_codes()"""
		return self._core.io.query_all_syst_errors(include_codes=False)

	def query_all_errors_with_codes(self) -> List[Tuple[int, str]] | None:
		"""Queries and clears all the errors from the instrument's error queue.
		The method returns list of tuples (code: int, message: str). If no error is detected, the return value is None.
		The process is: querying 'SYSTem:ERRor?' in a loop until the error queue is empty."""
		return self._core.io.query_all_syst_errors(include_codes=True)

	def reset(self, timeout: int = 0) -> None:
		"""SCPI command: *RST
		Sends *RST command + calls the clear_status().
		If you define timeout > 0, the VISA timeout is set to that value just for this method call."""
		self._core.io.reset(timeout)

	def self_test(self, timeout: int = None) -> Tuple[int, str]:
		"""SCPI command: *TST?
		Performs instrument's self-test.
		Returns tuple (code:int, message: str). Code 0 means the self-test passed.
		You can define the custom timeout in milliseconds. If you do not define it, the method uses default self-test timeout (usually 60 secs)."""
		return self._core.io.self_test(timeout)

	def is_connection_active(self) -> bool:
		"""Returns true, if the VISA connection is active and the communication with the instrument still works.
		WARNING!!! this method queries the session's VISA Timeout and additionally, queries the *IDN? from the instrument,
		hence affects the performance of your application when used regularly."""
		return self._core.io.is_connection_active()

	def reconnect(self, force_close: bool = False) -> bool:
		"""If the connection is not active, the method tries to reconnect to the device.
		If the connection is active, and force_close is False, the method does nothing.
		If the connection is active, and force_close is True, the method closes, and opens the session again.
		Returns True, if the reconnection has been performed."""
		return self._core.io.reconnect(force_close)

	@property
	def resource_name(self) -> str:
		"""Returns the resource name used in the constructor."""
		return self._core.io.resource_name

	@property
	def opc_timeout(self) -> int:
		"""Sets / returns timeout in milliseconds for all the operations that use OPC synchronization."""
		return self._core.io.opc_timeout

	@opc_timeout.setter
	def opc_timeout(self, value: int) -> None:
		"""Sets / returns timeout in milliseconds for all the operations that use OPC synchronization."""
		self._core.io.opc_timeout = value

	@property
	def visa_timeout(self) -> int:
		"""Sets / returns visa IO timeout in milliseconds."""
		return self._core.io.visa_timeout

	@visa_timeout.setter
	def visa_timeout(self, value) -> None:
		"""Sets / returns visa IO timeout in milliseconds."""
		self._core.io.visa_timeout = value

	@property
	def data_chunk_size(self) -> int:
		"""Returns max chunk size of one data block."""
		return self._core.io.data_chunk_size

	@data_chunk_size.setter
	def data_chunk_size(self, chunk_size: int) -> None:
		"""Sets the maximum size of one block transferred during write/read operations."""
		self._core.io.data_chunk_size = chunk_size

	@property
	def visa_manufacturer(self) -> str:
		"""Returns the manufacturer of the current VISA session."""
		return self._core.io.visa_manufacturer

	@property
	def opc_sync_query_mechanism(self) -> OpcSyncQueryMechanism:
		"""Returns the current setting of the OPC-Sync query mechanism."""
		return self._core.io.opc_sync_query_mechanism

	@opc_sync_query_mechanism.setter
	def opc_sync_query_mechanism(self, mechanism: OpcSyncQueryMechanism) -> None:
		"""Sets the current setting of the OPC-Sync query mechanism."""
		self._core.io.opc_sync_query_mechanism = mechanism

	def go_to_local(self, mixed_mode: bool = True) -> None:
		"""Puts the instrument into local state.
		By default, the method uses a mechanism to keep the instrument in a mixed mode: remote and local.
		That means, you can remote-control your instrument, and at the same time it still allows manual control.
		Set the mixed_mode to False, if you want your instrument to go to remote mode as soon as it receives the first remote command."""
		self._core.io.go_to_local(mixed_mode)

	def go_to_remote(self) -> None:
		"""Puts the instrument into remote state."""
		self._core.io.go_to_remote()

	def lock_resource(self, timeout: int, requested_key: str | bytes = None) -> bytes | str:
		"""Locks the instrument to prevent it from communicating with other clients."""
		return self._core.io.lock_resource(timeout, requested_key)

	def unlock_resource(self) -> None:
		"""Unlocks the instrument to other clients."""
		self._core.io.unlock_resource()

	def process_all_commands(self) -> None:
		"""SCPI command: *WAI
		Stops further commands processing until all commands sent before *WAI have been executed."""
		return self._core.io.write('*WAI')

	def write(self, cmd: str) -> None:
		"""Writes the command to the instrument as string.
		This method is an alias to the write_str() method."""
		self._core.io.write(cmd, log_info='Write')

	def write_str(self, cmd: str) -> None:
		"""Writes the command to the instrument as string.
		This method is an alias to write() method."""
		self._core.io.write(cmd, log_info='Write')

	def write_int(self, cmd: str, param: int) -> None:
		"""Writes the command to the instrument followed by the integer parameter:
		e.g.: cmd = 'SELECT:INPUT' param = '2', result command = 'SELECT:INPUT 2'"""
		self._core.io.write(f'{cmd} {param}', log_info='Write integer')

	def write_int_with_opc(self, cmd: str, param: int, timeout: int = None) -> None:
		"""Writes the command with OPC to the instrument followed by the integer parameter:
		e.g.: cmd = 'SELECT:INPUT' param = '2', result command = 'SELECT:INPUT 2'
		If you do not provide timeout, the method uses current opc_timeout."""
		self._core.io.write_with_opc(f'{cmd} {param}', timeout, log_info='Write integer with OPC')

	def write_float(self, cmd: str, param: float) -> None:
		"""Writes the command to the instrument followed by the boolean parameter:
		e.g.: cmd = 'CENTER:FREQ' param = '10E6', result command = 'CENTER:FREQ 10E6'"""
		self._core.io.write(f'{cmd} {Conv.float_to_str(param)}', log_info='Write float')

	def write_float_with_opc(self, cmd: str, param: float, timeout: int = None) -> None:
		"""Writes the command with OPC to the instrument followed by the boolean parameter:
		e.g.: cmd = 'CENTER:FREQ' param = '10E6', result command = 'CENTER:FREQ 10E6'
		If you do not provide timeout, the method uses current opc_timeout."""
		self._core.io.write_with_opc(f'{cmd} {Conv.float_to_str(param)}', timeout, log_info='Write float with OPC')

	def write_bool(self, cmd: str, param: bool) -> None:
		"""Writes the command to the instrument followed by the boolean parameter:
		e.g.: cmd = 'OUTPUT' param = 'True', result command = 'OUTPUT ON'"""
		self._core.io.write(f'{cmd} {Conv.bool_to_str(param)}', log_info='Write boolean')

	def write_bool_with_opc(self, cmd: str, param: bool, timeout: int = None) -> None:
		"""Writes the command with OPC to the instrument followed by the boolean parameter:
		e.g.: cmd = 'OUTPUT' param = 'True', result command = 'OUTPUT ON'
		If you do not provide timeout, the method uses current opc_timeout."""
		self._core.io.write_with_opc(f'{cmd} {Conv.bool_to_str(param)}', timeout, log_info='Write boolean with OPC')

	def query(self, query: str) -> str:
		"""Sends the string query to the instrument and returns the response as string.
		The response is trimmed of any trailing LF characters and has no length limit.
		This method is an alias to the query_str() method."""
		return self._core.io.query_str(query)

	def query_str(self, query: str) -> str:
		"""Sends the string query to the instrument and returns the response as string.
		The response is trimmed of any trailing LF characters and has no length limit.
		This method is an alias to the query() method."""
		return self._core.io.query_str(query)

	def query_stripped(self, query: str) -> str:
		"""Sends the string query to the instrument and returns the response as string stripped of the trailing LF and leading/trailing single/double quotes.
		The stripping of the leading/trailing quotes is blocked, if the string contains the quotes in the middle.
		"""
		return trim_str_response(self._core.io.query_str(query))

	def query_str_stripped(self, query: str) -> str:
		"""Sends the string query to the instrument and returns the response as string stripped of the trailing LF and leading/trailing single/double quotes.
		The stripping of the leading/trailing quotes is blocked, if the string contains the quotes in the middle.
		This method is an alias to the query_stripped() method."""
		return self.query_stripped(query)

	def query_bool(self, query: str) -> bool:
		"""Sends the query to the instrument and returns the response as boolean."""
		return self._core.io.query_bool(query)

	def query_int(self, query: str) -> int:
		"""Sends the query to the instrument and returns the response as integer."""
		return self._core.io.query_int(query)

	def query_float(self, query: str) -> float:
		"""Sends the query to the instrument and returns the response as float."""
		return self._core.io.query_float(query)

	def query_str_list(self, query: str, remove_blank_response: bool = False) -> List[str]:
		"""Sends the string query to the instrument and returns the response as List of strings, where the delimiter is comma (','). Each element of the list is trimmed for leading and trailing quotes.
		\nMeaning of the 'remove_blank_response':
		- False(default): whitespaces-only response is returned as a list with one empty element [''].
		- True: whitespaces-only response is returned as an empty list []."""
		return self._core.io.query_str_list(query, remove_blank_response)

	def query_bool_list(self, query: str) -> List[bool]:
		"""Sends the string query to the instrument and returns the response as List of booleans,
		where the delimiter is comma (',').
		Blank or empty response is returned as an empty list."""
		return self._core.io.query_bool_list(query)

	def write_str_with_opc(self, cmd: str, timeout: int = None) -> None:
		"""Writes the opc-synced command to the instrument.
		If you do not provide timeout, the method uses current opc_timeout."""
		self._core.io.write_with_opc(cmd, timeout)

	def write_with_opc(self, cmd: str, timeout: int = None) -> None:
		"""This method is an alias to the write_str_with_opc().
		Writes the opc-synced command to the instrument.
		If you do not provide timeout, the method uses current opc_timeout."""
		self._core.io.write_with_opc(cmd, timeout)

	def query_str_with_opc(self, query: str, timeout: int = None) -> str:
		"""Sends the opc-synced query to the instrument and returns the response as string.
		The response is trimmed of any trailing LF characters and has no length limit.
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_str_with_opc(query, timeout)

	def query_with_opc(self, query: str, timeout: int = None) -> str:
		"""This method is an alias to the write_str_with_opc().
		Sends the opc-synced query to the instrument and returns the response as string.
		The response is trimmed of any trailing LF characters and has no length limit.
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_str_with_opc(query, timeout)

	def query_str_list_with_opc(self, query: str, timeout: int = None, remove_blank_response: bool = False) -> List[str]:
		"""Sends a OPC-synced query and reads response from the instrument as csv-list.
		If you do not provide timeout, the method uses current opc_timeout.
		\nMeaning of the 'remove_blank_response':
		- False(default): whitespaces-only response is returned as a list with one empty element [''].
		- True: whitespaces-only response is returned as an empty list []."""
		return self._core.io.query_str_list_with_opc(query, timeout, remove_blank_response)

	def query_bool_with_opc(self, query: str, timeout: int = None) -> bool:
		"""Sends the opc-synced query to the instrument and returns the response as boolean.
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_bool_with_opc(query, timeout)

	def query_bool_list_with_opc(self, query: str, timeout: int = None) -> List[bool]:
		"""Sends a OPC-synced query and reads response from the instrument as csv-list of booleans.
		If you do not provide timeout, the method uses current opc_timeout.
		Blank or empty response is returned as an empty list."""
		return self._core.io.query_bool_list_with_opc(query, timeout)

	def query_int_with_opc(self, query: str, timeout: int = None) -> int:
		"""Sends the opc-synced query to the instrument and returns the response as integer.
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_int_with_opc(query, timeout)

	def query_float_with_opc(self, query: str, timeout: int = None) -> float:
		"""Sends the opc-synced query to the instrument and returns the response as float.
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_float_with_opc(query, timeout)

	def write_bin_block(self, cmd: str, payload: bytes) -> None:
		"""Writes all the payload as binary data block to the instrument.
		The binary data header is added at the beginning of the transmission automatically, do not include it in the payload!!!"""
		self._core.io.write_bin_block(cmd, payload)

	def query_bin_block(self, query: str) -> bytes:
		"""Queries binary data block to bytes.
		Throws an exception if the returned data was not a binary data.
		Returns data:bytes"""
		return self._core.io.query_bin_block(query)

	def query_bin_block_with_opc(self, query: str, timeout: int = None) -> bytes:
		"""Sends a OPC-synced query and returns binary data block to bytes.
		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_bin_block_with_opc(query, timeout)

	def query_bin_or_ascii_float_list(self, query: str) -> List[float]:
		"""Queries a list of floating-point numbers that can be returned as ASCII or binary format.

		* For ASCII format, the list numbers are decoded as comma-separated values.
		* For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32).
		"""
		return self._core.io.query_bin_or_ascii_float_list(query)

	def query_bin_or_ascii_float_list_with_opc(self, query: str, timeout: int = None) -> List[float]:
		"""Sends a OPC-synced query and reads a list of floating-point numbers that can be returned as ASCII or binary format.

		* For ASCII format, the list numbers are decoded as comma-separated values.
		* For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32).

		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_bin_or_ascii_float_list_with_opc(query, timeout)

	def query_bin_or_ascii_int_list(self, query: str) -> List[int]:
		"""Queries a list of floating-point numbers that can be returned as ASCII or binary format.

		* For ASCII format, the list numbers are decoded as comma-separated values.
		* For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32).
		"""
		return self._core.io.query_bin_or_ascii_int_list(query)

	def query_bin_or_ascii_int_list_with_opc(self, query: str, timeout: int = None) -> List[int]:
		"""Sends a OPC-synced query and reads a list of floating-point numbers that can be returned as ASCII or binary format.

		* For ASCII format, the list numbers are decoded as comma-separated values.
		* For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32).

		If you do not provide timeout, the method uses current opc_timeout."""
		return self._core.io.query_bin_or_ascii_int_list_with_opc(query, timeout)

	# Write / Read to file
	def query_bin_block_to_file(self, query: str, file_path: str, append: bool = False) -> None:
		"""Queries binary data block to the provided file.
		If append is False, any existing file content is discarded.
		If append is True, the new content is added to the end of the existing file, or if the file does not exit, it is created.
		Throws an exception if the returned data was not a binary data.
		Example for transferring a file from Instrument -> PC:
		query = f"MMEM:DATA? '{INSTR_FILE_PATH}'".

		Alternatively, use the dedicated methods for this purpose:

		* ``send_file_from_pc_to_instrument()``
		* ``read_file_from_instrument_to_pc()``

		"""
		self._core.io.query_bin_block_to_file(query, file_path, append)

	def query_bin_block_to_file_with_opc(self, query: str, file_path: str, append: bool = False, timeout: int = None) -> None:
		"""Sends a OPC-synced query and writes the returned data to the provided file.
		If append is False, any existing file content is discarded.
		If append is True, the new content is added to the end of the existing file, or if the file does not exit, it is created.
		Throws an exception if the returned data was not a binary data."""
		self._core.io.query_bin_block_to_file_with_opc(query, file_path, append, timeout)

	def write_bin_block_from_file(self, cmd: str, file_path: str) -> None:
		"""Writes data from the file as binary data block to the instrument using the provided command.
		Example for transferring a file from PC -> Instrument:
		cmd = f"MMEM:DATA '{INSTR_FILE_PATH}',".

		Alternatively, use the dedicated methods for this purpose:

		* ``send_file_from_pc_to_instrument()``
		* ``read_file_from_instrument_to_pc()``

		"""
		self._core.io.write_bin_block_from_file(cmd, file_path)

	def send_file_from_pc_to_instrument(self, source_pc_file: str, target_instr_file: str) -> None:
		"""SCPI Command: MMEM:DATA \n
		Sends file from PC to the instrument."""
		self._core.io.send_file_from_pc_to_instrument(source_pc_file, target_instr_file)

	def read_file_from_instrument_to_pc(self, source_instr_file: str, target_pc_file: str, append_to_pc_file: bool = False) -> None:
		"""SCPI Command: MMEM:DATA? \n
		Reads file from instrument to the PC. \n
		Set the ``append_to_pc_file`` to True if you want to append the read content to the end of the existing PC file."""
		self._core.io.read_file_from_instrument_to_pc(source_instr_file, target_pc_file, append_to_pc_file)

	def get_file_size(self, instr_file: str) -> int | None:
		"""Return size of the instrument file, or None if the file does not exist."""
		return self._core.io.get_file_size(instr_file)

	def file_exists(self, instr_file: str) -> bool:
		"""Returns true, if the instrument file exist."""
		return self.get_file_size(instr_file) is not None

	def get_last_sent_cmd(self) -> str:
		"""Returns the last commands sent to the instrument. Only works in simulation mode."""
		return self._core.get_last_sent_cmd()

	def get_lock(self) -> threading.RLock:
		"""Returns the thread lock for the current session. \n
		By default:

		* If you create a new RsInstrument instance with new VISA session, the session gets a new thread lock. You can assign it to another RsInstrument sessions in order to share one physical instrument with a multi-thread access.
		* If you create a new RsInstrument from an existing session, the thread lock is shared automatically making both instances multi-thread safe.

		You can always assign new thread lock by calling ``driver.utilities.assign_lock()``"""
		return self._core.io.get_lock()

	def assign_lock(self, lock: threading.RLock) -> None:
		"""Assigns the provided thread lock."""
		self._core.io.assign_lock(lock)

	def clear_lock(self) -> None:
		"""Clears the existing thread lock, making the current session thread-independent from others that might share the current thread lock."""
		self._core.io.clear_lock()

	def get_total_execution_time(self) -> timedelta:
		"""Returns total time spent by the library on communicating with the instrument.
		This time is always shorter than get_total_time(), since it does not include gaps between the communication.
		You can reset this counter with reset_time_statistics()."""
		return self._core.io.total_execution_time

	def get_total_time(self) -> timedelta:
		"""Returns delta time spent by the library between the get_total_time_startpoint() and now.
		This time is always longer than get_total_execution_time(), since it also includes all other activities besides the communication.
		You can set the total time startpoint to now with reset_time_statistics()."""
		return datetime.now() - self._core.io.total_time_startpoint

	def get_total_time_startpoint(self) -> datetime:
		"""Returns time from which the execution started.
		This is the value that the get_total_time() calculates as its reference.
		Calling the reset_time_statistics() sets this time to now."""
		return self._core.io.total_time_startpoint

	def reset_time_statistics(self) -> None:
		"""Resets all execution and total time counters.
		Affects the results of get_total_time(), get_total_execution_time() and get_total_time_startpoint()"""
		self._core.io.reset_time_statistics()

	@staticmethod
	def assert_minimum_version(min_version: str) -> None:
		"""Asserts that the driver version fulfills the minimum required version you have entered.
		This way you make sure your installed driver is of the entered version or newer."""
		min_version_list = min_version.split('.')
		curr_version_list = RsInstrument._driver_version_const.split('.')
		count_min = len(min_version_list)
		count_curr = len(curr_version_list)
		count = count_min if count_min < count_curr else count_curr
		for i in range(count):
			minimum = int(min_version_list[i])
			curr = int(curr_version_list[i])
			if curr > minimum:
				break
			if curr < minimum:
				raise Exception(f"Assertion for minimum RsInstrument version failed. Current version: '{RsInstrument._driver_version_const}', minimum required version: '{min_version}'")

	def instr_err_suppressor(self, visa_tout_ms: int = 0, suppress_only_codes: int | List[int] = None) -> InstrErrorSuppressor:
		"""Returns Context Manager that suppresses the instrument errors.
		Other exceptions types are still raised.
		On entering the context, this class clears all the instrument status errors.
		:param visa_tout_ms: VISA Timeout in milliseconds, that is set for this context. Afterward, it is changed back. Default value: do-not-change.
		:param suppress_only_codes: You can enter a code or list of codes for errors to be suppressed. Other errors will be reported. Example: If you enter -113 here, only the 'Undefined Header' error will be suppressed. Default value: suppress-all-errors."""
		return InstrErrorSuppressor(self._core.io, visa_tout_ms, suppress_only_codes)

	def visa_tout_suppressor(self, visa_tout_ms: int = 0) -> VisaTimeoutSuppressor:
		"""Returns Context Manager that suppresses the VISA timeout error.
		Careful!!!: Only the very first VISA Timeout exception is suppressed,
		and afterward the context ends. Therefore, use only one command per context manager,
		if you do not want to skip the following ones.
		:param visa_tout_ms: VISA Timeout in milliseconds, that is set for this context. Afterward, it is changed back. Default value: do-not-change.
		"""
		return VisaTimeoutSuppressor(self._core.io, visa_tout_ms)

	@property
	def instrument_options(self) -> List[str]:
		"""Returns all the instrument options.
		The options are sorted in the ascending order starting with K-options and continuing with B-options"""
		return self._core.io.instr_options.get_all()

	def has_instr_option(self, options: str | List[str]) -> bool:
		"""Returns true, if the entered options (case-insensitive) matches at least one of the installed options (or-logic).
		You can enter either a string with one option, or more options '/'-separated, or more options as a list of strings.
		If K0 is present, all the K-options are reported as present. B-options are not affected by K0.
		Example 1: options='k23' returns true, if the instrument has the option 'K23'.
		Example 2: options='k23 / K23e' returns true, if the instrument has either the option 'K23' or the option 'K23E'.
		Example 3: options=['k11','K22'] returns true, if the instrument has either the option 'K11' or the option 'K22'."""
		return self._core.io.instr_options.has(options)

	def has_instr_option_regex(self, re_options: str | List[str]) -> bool:
		"""Returns true, if the entered regex string (case-insensitive) matches at least one of the installed options.
		The match must be complete, not just partial (search).
		You can enter either a string with one option, or more options '/'-separated, or more options as a list of strings.
		Example 1: re_options='k10.' returns true, if the instrument contains any option 'K100' ... up to 'K109' .
		Example 2: re_options='k10. / k20.*' returns true, if the instrument contains any of the options 'K10x' or 'K20xxx'.
		Example 3: re_options=['k10.', 'k20.*'] returns true, if the instrument contains any options 'K10x' or 'K20xxx'."""
		return self._core.io.instr_options.has_regex(re_options)

	def has_instr_option_k0(self) -> bool:
		"""Returns true, if the instrument has K0 installed."""
		return self._core.io.instr_options.has_k0()

	def add_instr_option(self, option: str) -> None:
		"""Adds new option if not already existing."""
		self._core.io.instr_options.add(option)

	def remove_instr_option(self, option: str) -> None:
		"""Removes the option if exists."""
		self._core.io.instr_options.remove(option)



"""VISA communication interface for SCPI-based instrument remote control.
:version: 1.124.0
:copyright: 2025 by Rohde & Schwarz GMBH & Co. KG
:license: MIT, see LICENSE for more details.
"""

__version__ = '1.124.0'

# Main class
from RsInstrument.RsInstrument import RsInstrument

# Bin data format
from RsInstrument.RsInstrument import BinFloatFormat, BinIntFormat

# Opc-Sync Query Mechanism
from RsInstrument.Internal.InstrumentSettings import OpcSyncQueryMechanism

# Exceptions
from RsInstrument.Internal.InstrumentErrors import (
    RsInstrException,
    TimeoutException,
    StatusException,
    UnexpectedResponseException,
    ResourceError,
    DriverValueError
)

# Callback Event Argument prototypes
from RsInstrument.Internal.IoTransferEventArgs import IoTransferEventArgs

# SCPI Logger
from RsInstrument.Internal.ScpiLogger import LoggingMode

# Utilities
from RsInstrument.Internal.Utilities import (
    size_to_kb_mb_gb_string,
    size_to_kb_mb_string,
    value_to_si_string
)


__all__ = [
    "RsInstrument",
    "BinIntFormat",
    "BinFloatFormat",
    "TimeoutException",
    "IoTransferEventArgs",
    "LoggingMode",
    "StatusException",
    "ResourceError",
    "size_to_kb_mb_string",
    "size_to_kb_mb_gb_string",
    "value_to_si_string",
    "RsInstrException",
    "OpcSyncQueryMechanism",
    "UnexpectedResponseException",
    "DriverValueError",
]



"""Contains definition for an argument that is suppressed and not exposed to the user.
Usually such arguments are also linked to a callback."""

from .Types import DataType


class ArgSingleSuppressed(object):
	"""Single suppressed Argument - used in Query_XxXx_Suppressed() to remove it from the returned value.
	It does not contain:
	- 'value' attribute, since this is discarded or linked internally directly  in the Query_XxXx_Suppressed().
	- 'is_optional' attribute, since it is always mandatory."""

	def __init__(self, argument_ix: int, data_type: DataType, is_open_list: bool = False, repetition: int = 1, intern_link: str = None):
		self.name = ''
		self.argument_ix = argument_ix
		self.data_type = data_type
		self.is_open_list = is_open_list
		self.repetition = repetition
		self.intern_link = intern_link

	def __str__(self):
		out = f"{self.data_type.name} '{self.name}' SuppressedArg"
		if self.is_open_list is False and self.repetition > 1:
			out += f' [{self.repetition}]'
		elif self.is_open_list:
			out += f' [{self.repetition}...]'

		if self.intern_link:
			out += f", Linking: '{self.intern_link}'"

		return out



"""Provides conversions between SCPI enums and strings."""

from enum import Enum
from typing import List
from .Utilities import trim_str_response

enum_spec_prefixes = {'_minus': '-', '_plus': '+', '_': ''}
enum_spec_strings = {'_dash_': '-', '_dot_': '.', '_cma_': ','}


class ScpiEnum:
    """Provides methods for enum members search including the ones with custom SCPI strings."""

    def __init__(self, enum_type):
        self.enum_type = enum_type
        self._members_raw = [x.name for x in self.enum_type]
        self.has_custom_values = any(isinstance(x.value, str) for x in self.enum_type)
        self.has_quotes = False
        self._custom_values = None
        self._members_special = None
        self._members_stripped_quotes = None
        if self.has_custom_values:
            self.members = []
            self._custom_values = {}
            for x in self.enum_type:
                item = x.value if isinstance(x.value, str) else x.name
                if not self.has_quotes:
                    self.has_quotes = ("'" in item or '"' in item)
                self.members.append(item)
                self._custom_values[x.name] = item
        else:
            self.members = self._members_raw

    def get_scpi_value(self, enum_value: str) -> str:
        """Returns the SCPI value of the enum item: name for the integer value and value for the string value."""
        if not self.has_custom_values:
            return enum_value
        return self._custom_values[enum_value]

    def find_in_enum_members(self, item: str, force_comma_remove: bool) -> Enum | None:
        """Returns either an EnumMember item or null if not found.
        The matching is done against the Members, and if unsuccessful, then against the _members_special
        The response is always a _members_raw item or None."""
        ix = self._find_ix_in_enum_members(item, force_comma_remove, self.members)
        if ix >= 0:
            return self.enum_type[self._members_raw[ix]]

        self._init_special_values()
        ix = self._find_ix_in_enum_members(item, force_comma_remove, self._members_special)
        if ix >= 0:
            return self.enum_type[self._members_raw[ix]]

        if self.has_quotes:
            self._init_stripped_quotes()
            item = trim_str_response(item)
            ix = self._find_ix_in_enum_members(item, force_comma_remove, self._members_stripped_quotes)
            if ix >= 0:
                return self.enum_type[self._members_raw[ix]]

        return None

    @staticmethod
    def _find_ix_in_enum_members(item: str, force_comma_remove: bool, enum_members: List[str]) -> int:
        """Matches an item in the provided list of enum_member strings.
        The item must be not fully matched.
        The item is matched if a member string starts with the item (the item is a prefix of the member).
        Example: item='CONN' matches the enum_member 'CONNected'.
        If the item contains a comma, the function checks if there is a comma defined in the enum_members.
        - If no, the comma and all after it is removed.
        - If yes, the comma is kept.
        You can override the behavior by forcing the removal of the comma
        Returns found index in the enum_members list."""
        if ',' in item:
            trim_comma = True
            if not force_comma_remove:
                for x in enum_members:
                    if '_cma_' in x or ',' in x:
                        trim_comma = False
                        break
            if trim_comma:
                item = item[:item.index(',')].strip()
        i = 0
        for x in enum_members:
            if x.startswith(item):
                return i
            i += 1

        # smart matching:
        # item = 'MAX' matches enum_member 'MAXpeak'
        # item = 'SPECtrum1' matches enum_member 'SPEC1'
        # item = 'SPEC' matches enum_member 'SPECtrum1'

        item = ''.join([c for c in item if not c.islower()])
        # item must be longer than 1 character
        if len(item) < 2:
            return -1
        i = 0
        for x in enum_members:
            x_uc = ''.join([c for c in x if not c.islower()])
            if x_uc == item:
                return i
            i += 1

        return -1

    def _init_special_values(self) -> None:
        """Convert the members to the SCPI values - values to be sent to the instrument
        Resolves escapes:
        '^_' => ''
        '^_minus' => '-'
        '_dash_' => '-'
        '_dot_' => '.'
        '_cma_' => ','
        This is a lazy init of the self._members_special property, since most of the enums do not have any special values.
        """
        if self._members_special is not None:
            return
        self._members_special = []
        for i in range(len(self.members)):
            mem = self.members[i]
            for key in enum_spec_prefixes:
                if mem.startswith(key):
                    mem = enum_spec_prefixes[key] + mem[len(key):]
            for key in enum_spec_strings:
                mem = mem.replace(key, enum_spec_strings[key])
            self._members_special.append(mem)

    def _init_stripped_quotes(self) -> None:
        """Convert the members to the SCPI values without the quotes."""
        if self._members_stripped_quotes is not None:
            return
        self._members_stripped_quotes = []
        for i in range(len(self.members)):
            self._members_stripped_quotes.append(trim_str_response(self.members[i]))



"""Single argument definition for a scalar argument."""

from .ConverterFromScpiString import ConverterFromScpiString
from .ConverterToScpiString import ConverterToScpiString
from .InstrumentErrors import RsInstrException

from .Types import DataType


class ArgSingle(object):
	"""Single Argument outside a structure - used for composing query arguments.
	Contains the argument value as well (self.value)."""

	def __init__(self, name: str, value, data_type: DataType, enum_type=None, is_optional: bool = False, is_open_list: bool = False, repetition: int = 1, intern_link: str = None):
		self.name = name if name else ''
		self.argument_ix = None
		self.value = value
		self.data_type = data_type
		self.enum_type = enum_type
		self.is_optional = is_optional
		self.is_open_list = is_open_list
		self.repetition = repetition
		self.intern_link = intern_link
		self.conv_from_scpi_string = None
		self.conv_to_scpi_string = None

		if self.data_type.is_scalar_enum:
			# self.assert_mandatory_has_value(self)
			self.conv_from_scpi_string = ConverterFromScpiString(self.data_type, self.enum_type)
			self.conv_to_scpi_string = ConverterToScpiString(self.data_type, self.enum_type)
		elif self.data_type.is_list_enum:
			# self.assert_mandatory_has_value(self)
			if self.value is not None:
				self.conv_from_scpi_string = ConverterFromScpiString(self.data_type, self.enum_type)
				self.conv_to_scpi_string = ConverterToScpiString(self.data_type, self.enum_type)
		else:
			self.conv_from_scpi_string = ConverterFromScpiString(self.data_type)
			self.conv_to_scpi_string = ConverterToScpiString(self.data_type)

		self.check_consistency()

	@classmethod
	def as_open_list(cls, name: str, value: object, data_type: DataType, enum_type=None) -> 'ArgSingle':
		"""Creates new ArgSingle of open list type.Use this method for all non-interleaved list types. \n
		:param name: name of the argument
		:param value: value of the argument
		:param data_type: data type of the argument
		:param enum_type: enum type if the data_type is Enum or EnumExt (or list of those)
		:return: ArgSingle object of an open list type"""
		return cls(name, value, data_type, enum_type, False, True, 1, None)

	def __str__(self):
		opt = '~' if self.is_optional else ''
		name = f" '{self.name}'" if self.name != '' else ''
		out = f"SingleArg {opt}{self.data_type.name}{name}"

		if self.is_open_list is False and self.repetition > 1:
			out += f' [{self.repetition}]'
		elif self.is_open_list:
			out += f' [{self.repetition}...]'
		if self.intern_link:
			out += f", Linking: '{self.intern_link}'"
		if self.value:
			out += f', value: {self.value}'
		else:
			out += ", <no value>"
		if self.intern_link:
			out += f", Linking: '{self.intern_link}'"
		return out

	# noinspection PyUnusedLocal
	def has_value(self, value_obj=None) -> bool:
		"""Returns true, if the argument has value."""
		return self.value is not None

	# noinspection PyUnusedLocal
	def assert_is_optional(self, obj=None) -> None:
		"""Asserts that the parameter is optional.
		If not, the method throws an exception."""
		if self.is_optional:
			return
		raise RsInstrException(f'Single argument is not optional: {self}')

	# noinspection PyUnusedLocal
	def assert_mandatory_has_value(self, value_obj=None) -> None:
		"""Asserts that if the parameter is mandatory, it must have value assigned.
		If not, the method throws an exception."""
		if self.is_optional:
			return
		if self.value is None:
			raise ValueError(f'Mandatory single argument has no value: {self}')

	def set_scalar_value_from_str(self, string: str) -> None:
		"""Sets scalar value from input string"""
		self.value = self.conv_from_scpi_string.get_value(string)

	def check_consistency(self) -> None:
		"""Checks the consistency of the object"""
		if self.value is None:
			return
		if isinstance(self.value, list):
			if self.data_type.is_scalar:
				raise RsInstrException(f'Argument real data type is list, but it is declared as {self.data_type}. Value: {self.value}')
		else:
			if self.data_type.is_list:
				raise RsInstrException(f'Argument real data type is scalar, but it is declared as {self.data_type}. Value: {self.value}')



"""Definition for an argument which is a part of a structure."""

from .ConverterFromScpiString import ConverterFromScpiString
from .ConverterToScpiString import ConverterToScpiString
from .Types import DataType
from .InstrumentErrors import RsInstrException


class ArgStruct(object):
	"""Describes an argument in data structures.
	This info is used to parse a string query response to the output structure,
	or to parse the output structure to the string parameter for writing.
	Contains reference to the value in the owning structure."""

	def __init__(self, name: str, data_type: DataType, enum_type=None, is_optional=False, is_open_list=False, repetition=1, intern_link: str = None):
		self.argument_ix = None
		self.name = name
		self.data_type = data_type
		self.is_optional = is_optional
		self.is_open_list = is_open_list
		self.repetition = repetition
		self.intern_link = intern_link
		self.enum_type = enum_type
		self.conv_from_scpi_string = ConverterFromScpiString(self.data_type, self.enum_type)
		self.conv_to_scpi_string = ConverterToScpiString(self.data_type, self.enum_type)

	@classmethod
	def scalar_int(cls, name: str, intern_link: str = None):
		"""Describes mandatory scalar integer argument."""
		return cls(name, DataType.Integer, None, False, False, 1, intern_link)

	@classmethod
	def scalar_int_ext(cls, name: str, intern_link: str = None):
		"""Describes mandatory scalar extended integer argument."""
		return cls(name, DataType.IntegerExt, None, False, False, 1, intern_link)

	@classmethod
	def scalar_str(cls, name: str, intern_link: str = None):
		"""Describes mandatory scalar string argument."""
		return cls(name, DataType.String, None, False, False, 1, intern_link)

	@classmethod
	def scalar_raw_str(cls, name: str, intern_link: str = None):
		"""Describes mandatory scalar raw string argument."""
		return cls(name, DataType.RawString, None, False, False, 1, intern_link)

	@classmethod
	def scalar_bool(cls, name: str, intern_link: str = None):
		"""Describes mandatory scalar boolean argument."""
		return cls(name, DataType.Boolean, None, False, False, 1, intern_link)

	@classmethod
	def scalar_float(cls, name: str, intern_link: str = None):
		"""Describes mandatory scalar float argument."""
		return cls(name, DataType.Float, None, False, False, 1, intern_link)

	@classmethod
	def scalar_float_ext(cls, name: str, intern_link: str = None):
		"""Describes mandatory scalar extended float argument."""
		return cls(name, DataType.FloatExt, None, False, False, 1, intern_link)

	@classmethod
	def scalar_enum(cls, name: str, enum_type, intern_link: str = None):
		"""Describes mandatory scalar float argument."""
		return cls(name, DataType.Enum, enum_type, False, False, 1, intern_link)

	@classmethod
	def scalar_str_optional(cls, name: str, intern_link: str = None):
		"""Describes optional scalar string argument."""
		return cls(name, DataType.String, None, True, False, 1, intern_link)

	@classmethod
	def scalar_raw_str_optional(cls, name: str, intern_link: str = None):
		"""Describes optional scalar raw string argument."""
		return cls(name, DataType.RawString, None, True, False, 1, intern_link)

	@classmethod
	def scalar_bool_optional(cls, name: str, intern_link: str = None):
		"""Describes optional scalar boolean argument."""
		return cls(name, DataType.Boolean, None, True, False, 1, intern_link)

	@classmethod
	def scalar_int_optional(cls, name: str, intern_link: str = None):
		"""Describes optional scalar integer argument."""
		return cls(name, DataType.Integer, None, True, False, 1, intern_link)

	@classmethod
	def scalar_int_ext_optional(cls, name: str, intern_link: str = None):
		"""Describes optional scalar extended integer argument."""
		return cls(name, DataType.IntegerExt, None, True, False, 1, intern_link)

	@classmethod
	def scalar_float_optional(cls, name: str, intern_link: str = None):
		"""Describes optional scalar float argument."""
		return cls(name, DataType.Float, None, True, False, 1, intern_link)

	@classmethod
	def scalar_float_ext_optional(cls, name: str, intern_link: str = None):
		"""Describes optional scalar extended float argument."""
		return cls(name, DataType.FloatExt, None, True, False, 1, intern_link)

	@classmethod
	def scalar_enum_optional(cls, name: str, enum_type, intern_link: str = None):
		"""Describes optional scalar float argument."""
		return cls(name, DataType.Enum, enum_type, True, False, 1, intern_link)

	def __str__(self):
		opt = '~' if self.is_optional else ''
		name = f" '{self.name}'" if self.name != '' else ''
		out = f"StructArg {opt}{self.data_type.name}{name}"

		if self.is_open_list is False and self.repetition > 1:
			out += f' [{self.repetition}]'
		elif self.is_open_list:
			out += f' [{self.repetition}...]'
		if self.intern_link:
			out += f", Linking: '{self.intern_link}'"
		return out

	def has_value(self, obj) -> bool:
		"""Returns True, if the entered object attribute has value."""
		return getattr(obj, self.name) is not None

	def assert_is_optional(self, obj) -> None:
		"""Asserts that the parameter is optional.
		If not, the method throws an exception."""
		if self.is_optional:
			return
		value = getattr(obj, self.name)
		if value is None:
			raise RsInstrException(f"Structure '{obj}', argument without value is not optional: {self}")
		else:
			raise RsInstrException(f"Structure '{obj}', argument is not optional: {self}', value '{value}'")

	def assert_mandatory_has_value(self, obj) -> None:
		"""Asserts that if the parameter is mandatory, it must have value assigned.
		If not, the method throws an exception."""
		if self.is_optional:
			return
		if getattr(obj, self.name) is None:
			raise ValueError(f"Mandatory structure '{obj}' argument '{self.name}' has no value.")



"""Tools for calling an instrument via SCPI commands."""

import argparse
import functools
import json
import logging
import sys
import typing

try:
    # noinspection PyUnusedImports
    from fastmcp import FastMCP
    # noinspection PyUnusedImports
    from starlette.responses import JSONResponse

    MCP_INSTALLED = True
except ImportError:
    MCP_INSTALLED = False

# noinspection PyProtectedMember
from RsInstrument import RsInstrument, __version__

logger = logging.getLogger(__name__)


def safe_tool(fn):
    """Decorator converting any raised Exception to standardized 'Error: <msg>' string.

    Applied to MCP-exposed functions so they never raise.
    """

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        """Wrapper function catching exceptions."""
        try:
            return fn(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-except
            logger.error("Exception in tool %s: %s", fn.__name__, e)
            return f"Error: {e}"

    return wrapper


@safe_tool
def instrument_query_scpi(command: str, resource: str, opc_timeout: int = 5000) -> str:
    """Query a command from an instrument via RsInstrument.

    Args:
        command: The SCPI query command to send to the instrument.
        resource: The VISA resource string of the instrument.
        opc_timeout: Timeout in milliseconds for the operation complete (OPC) query.
            Default is 5000 ms.

    Returns:
        The response from the instrument.
    """
    with RsInstrument(resource) as inst:
        inst.opc_timeout = opc_timeout
        response = inst.query(command)
        return response.strip()


@safe_tool
def instrument_write_scpi(command: str, resource: str, opc_timeout: int = 5000) -> str:
    """Write a command to an instrument via RsInstrument.

    Args:
        command: The SCPI write command to send to the instrument.
        resource: The VISA resource string of the instrument.
        opc_timeout: Timeout in milliseconds for the operation complete (OPC) query.
            Default is 5000 ms.

    Returns:
        The response from the instrument.
    """
    with RsInstrument(resource) as inst:
        inst.opc_timeout = opc_timeout
        inst.write(command)
        return "Write command executed successfully."


@safe_tool
def instrument_fetch_errors(resource: str, opc_timeout: int = 5000) -> str:
    """Fetch errors from an instrument via RsInstrument.

    Args:
        resource: The VISA resource string of the instrument.
        opc_timeout: Timeout in milliseconds for the operation complete (OPC) query.
            Default is 5000 ms.

    Returns:
        The response from the instrument.
    """
    with RsInstrument(resource) as inst:
        inst.opc_timeout = opc_timeout
        errors = inst.query_all_errors_with_codes()
        if not errors:
            return "No errors."

        return json.dumps([{"code": code, "message": msg} for code, msg in errors])


@safe_tool
def instrument_reset(resource: str, opc_timeout: int = 5000) -> str:
    """Reset an instrument via RsInstrument.

    Args:
        resource: The VISA resource string of the instrument.
        opc_timeout: Timeout in milliseconds for the operation complete (OPC) query.
            Default is 5000 ms.

    Returns:
        The response from the instrument.
    """
    with RsInstrument(resource) as inst:
        inst.opc_timeout = opc_timeout
        inst.reset()
        return "Instrument reset successfully."


def create_fastmcp_server(
    *args,
    tools: typing.Sequence[typing.Tuple[str, str, typing.Callable]] | None = None,
    health_endpoint: str = "/healthz",
    **kwargs,
):
    """Create a FastMCP tool for SCPI commands.

    Args:
        *args: Positional arguments to pass to FastMCP.
        tools: Register a sequence of tuples containing tool names, descriptions and their corresponding callables.
        health_endpoint: The FastMCP health endpoint.
        **kwargs: Keyword arguments to pass to FastMCP.
    """
    if not MCP_INSTALLED:
        raise ImportError(
            "mcp is required for this module. Please install with 'pip install RsInstrument[mcp]'"
        )
    if tools is None:
        tools = []
    name = f"{__package__}-mcp"
    kwargs.setdefault("name", name)
    fastmcp = FastMCP(*args, **kwargs)  # type: ignore

    @fastmcp.custom_route(health_endpoint, methods=["GET"])
    async def health_check(_):
        """Health check endpoint."""
        return JSONResponse(  # type: ignore
            {"status": "healthy", "service": name, "version": __version__}
        )

    fastmcp.tool(
        name="Instrument-Query-SCPI",
        description="Query a command from an instrument via RsInstrument.",
    )(instrument_query_scpi)
    fastmcp.tool(
        name="Instrument-Write-SCPI",
        description="Write a command to an instrument via RsInstrument.",
    )(instrument_write_scpi)
    fastmcp.tool(
        name="Instrument-Fetch-Errors",
        description="Fetch errors from an instrument via RsInstrument.",
    )(instrument_fetch_errors)
    fastmcp.tool(name="Instrument-Reset", description="Reset the instrument")(
        instrument_reset
    )
    for tool_name, tool_description, tool_callable in tools:
        fastmcp.tool(name=tool_name, description=tool_description)(tool_callable)
    return fastmcp


def run(
    *args,
    host: str = "localhost",
    port: int = 8000,
    transport: typing.Literal["stdio", "sse", "streamable-http"] = "stdio",
    tools: typing.Sequence[typing.Tuple[str, str, typing.Callable]] | None = None,
    health_endpoint: str = "/healthz",
    show_fastmcp_banner: bool = False,
    **kwargs,
):
    """Run the MCP server.

    Args:
        *args: Positional arguments to pass to FastMCP.
        host: The FastMCP Hostname/IP to bind.
        port: The FastMCP port to bind.
        transport: The FastMCP transport protocol to use. Options are 'stdio', 'sse', and 'streamable-http'.
        tools: Register a sequence of tuples containing tool names, descriptions and their corresponding callables.
        health_endpoint: The FastMCP health endpoint.
        show_fastmcp_banner: Whether to show the FastMCP banner on startup. Defaults to False.
        **kwargs: Keyword arguments to pass to FastMCP.
    """
    mcp = create_fastmcp_server(*args, tools=tools, health_endpoint=health_endpoint, **kwargs)
    if transport != "stdio":
        run_kwargs = {"host": host, "port": port}
    else:
        run_kwargs = {}
    if not show_fastmcp_banner:
        logger.info("Starting RsInstrument-mcp server...")
    mcp.run(transport=transport, show_banner=show_fastmcp_banner, **run_kwargs)


def create_parser() -> argparse.ArgumentParser:
    """Create command line argument parser."""
    parser = argparse.ArgumentParser(description="Run the RsInstrument MCP server.")
    parser.add_argument(
        "-V",
        "--version",
        help="Show version number and exit",
        action="version",
        version=__version__,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        help="Increase output (Option is additive to increase verbosity)",
        action="count",
        default=0,
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        help="Reduce output (Option is additive to decrease verbosity)",
        action="count",
        default=0,
    )
    parser.add_argument(
        "--host",
        dest="host",
        default="localhost",
        help="FastMCP Hostname/IP to bind (default: %(default)s)",
    )
    parser.add_argument(
        "--port",
        dest="port",
        type=int,
        default=8000,
        help="FastMCP port to bind (default: %(default)s)",
    )
    parser.add_argument(
        "--transport",
        dest="transport",
        type=str,
        choices=["stdio", "sse", "streamable-http"],
        default="streamable-http",
        help="FastMCP transport protocol (default: %(default)s)",
    )
    parser.add_argument(
        "--health-endpoint",
        dest="health_endpoint",
        type=str,
        default="/healthz",
        help="FastMCP health endpoint (default: %(default)s)",
    )
    parser.add_argument(
        "--show-fastmcp-banner",
        dest="show_fastmcp_banner",
        action="store_true",
        help="Show FastMCP banner on startup",
    )
    return parser


def main(argv: typing.Sequence[str] | None = None):
    """Main entry point for command line execution."""
    args = create_parser().parse_args(argv)

    logging.basicConfig(
        format="{asctime} [{levelname:^8}] ({filename}:{lineno}) {message}",
        datefmt="%Y-%m-%d %H:%M:%S",
        style="{",
    )

    default_log_level = logging.WARNING
    verbosity = default_log_level - ((args.verbose - args.quiet) * 10)
    log_level = min(logging.CRITICAL, max(logging.DEBUG, verbosity))
    logger.setLevel(log_level)
    try:
        run(
            transport=args.transport,
            host=args.host,
            port=args.port,
            health_endpoint=args.health_endpoint,
            show_fastmcp_banner=args.show_fastmcp_banner,
        )
    except (
        Exception
    ) as error:  # pragma: no cover - exercised in tests with forced exception
        if verbosity < default_log_level or default_log_level <= logging.DEBUG:
            logger.exception(error)
        else:
            logger.error(error)
            logger.warning("Hint: Rerun with '--verbose' to show exception traceback.")
        sys.exit(1)
    except KeyboardInterrupt:  # pragma: no cover - exercised in tests
        logger.warning("Aborted by user")
        sys.exit(1)


if __name__ == "__main__":
    main()
